"""
Interpreter for PDF content streams.
"""

import logging
from copy import copy
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Tuple,
    Type,
    Union,
)

from playa.color import PREDEFINED_COLORSPACE, ColorSpace, get_colorspace
from playa.content import (
    ContentObject,
    DashPattern,
    GraphicState,
    ImageObject,
    MarkedContent,
    PathObject,
    PathOperator,
    PathSegment,
    TagObject,
    TextObject,
    XObjectObject,
)
from playa.font import Font
from playa.parser import KWD, ContentParser, InlineImage, PDFObject
from playa.pdftypes import (
    LIT,
    MATRIX_IDENTITY,
    ContentStream,
    Matrix,
    ObjRef,
    Point,
    PSKeyword,
    PSLiteral,
    Rect,
    bool_value,
    dict_value,
    int_value,
    list_value,
    literal_name,
    num_value,
    resolve1,
    stream_value,
)
from playa.utils import mult_matrix
from playa.worker import _deref_document

if TYPE_CHECKING:
    from playa.document import Document
    from playa.page import Page

log = logging.getLogger(__name__)

# some predefined literals and keywords.
LITERAL_PAGE = LIT("Page")
LITERAL_PAGES = LIT("Pages")
LITERAL_FORM = LIT("Form")
LITERAL_IMAGE = LIT("Image")
TextSeq = Iterable[Union[int, float, bytes]]


def make_seg(operator: PathOperator, *points: Point):
    return PathSegment(operator, points)


@dataclass
class TextState:
    """Mutable text state (not for public consumption).

    Exceptionally, the line matrix and text matrix are represented
    more compactly with the line matrix itself in `line_matrix`, which
    gets translated by `glyph_offset` for the current glyph (note:
    expressed in **user space**), which pdfminer confusingly called
    `linematrix`, to produce the text matrix.

    Attributes:
      line_matrix: The text line matrix, which defines (in user
        space) the start of the current line of text, which may or may
        not correspond to an actual line because PDF is a presentation
        format.
      glyph_offset: The offset of the current glyph with relation to
        the line matrix, in text space units.
    """

    line_matrix: Matrix = MATRIX_IDENTITY
    glyph_offset: Point = (0, 0)

    def reset(self) -> None:
        """Reset the text state"""
        self.line_matrix = MATRIX_IDENTITY
        self.glyph_offset = (0, 0)


class FontMapping(Mapping[str, Font]):
    """Lazy mapping of font names to fonts in a resource dictionary."""

    def __init__(self, mapping: PDFObject, doc: "Document") -> None:
        self._fontmap: Dict[str, Font] = {}
        self._doc = doc
        try:
            self._mapping: Dict[str, PDFObject] = dict_value(mapping)
        except TypeError:
            log.warning("Font mapping not a dict: %r", mapping)
            self._mapping = {}

    def __len__(self) -> int:
        return len(self._mapping)

    def __iter__(self) -> Iterator[str]:
        return iter(self._mapping)

    def __getitem__(self, fontid: str) -> Font:
        if fontid in self._fontmap:
            return self._fontmap[fontid]
        spec = self._mapping[fontid]
        objid = 0  # Not a valid object ID (will not cache)
        if isinstance(spec, ObjRef):
            objid = spec.objid
        try:
            self._fontmap[fontid] = self._doc.get_font(objid, dict_value(spec))
        except Exception:
            log.warning(
                "Invalid font dictionary for Font %r: %r",
                fontid,
                spec,
                exc_info=True,
            )
            self._fontmap[fontid] = self._doc.get_font(objid, None)
        return self._fontmap[fontid]


class LazyInterpreter(Iterator[ContentObject]):
    """Interpret the page yielding lazy objects."""

    ctm: Matrix

    def __init__(
        self,
        page: "Page",
        contents: Iterable[PDFObject],
        resources: Union[Dict, None] = None,
        ctm: Union[Matrix, None] = None,
        gstate: Union[GraphicState, None] = None,
        parent_key: Union[int, None] = None,
        ignore_colours: bool = False,
        filter_classes: Union[Collection[Type[ContentObject]], None] = None,
        restrict_ops: Union[Collection[PSKeyword], None] = None,
    ) -> None:
        self._dispatch: Dict[PSKeyword, Tuple[Callable, int]] = {
            KWD(b"B"): (self.do_B, 0),
            KWD(b"BDC"): (self.do_BDC, 2),
            KWD(b"BI"): (self.do_BI, 0),
            KWD(b"BMC"): (self.do_BMC, 1),
            KWD(b"BT"): (self.do_BT, 0),
            KWD(b"BX"): (self.do_BX, 0),
            KWD(b"B*"): (self.do_B_a, 0),
            KWD(b"CS"): (self.do_CS, 1),
            KWD(b"DP"): (self.do_DP, 2),
            KWD(b"Do"): (self.do_Do, 1),
            KWD(b"EI"): (self.do_EI, 1),
            KWD(b"EMC"): (self.do_EMC, 0),
            KWD(b"ET"): (self.do_ET, 0),
            KWD(b"EX"): (self.do_EX, 0),
            KWD(b"F"): (self.do_F, 0),
            KWD(b"G"): (self.do_G, 1),
            KWD(b"ID"): (self.do_ID, 0),
            KWD(b"J"): (self.do_J, 1),
            KWD(b"K"): (self.do_K, 4),
            KWD(b"M"): (self.do_M, 1),
            KWD(b"MP"): (self.do_MP, 1),
            KWD(b"Q"): (self.do_Q, 0),
            KWD(b"RG"): (self.do_RG, 3),
            KWD(b"S"): (self.do_S, 0),
            KWD(b"SC"): (self.do_SC, 0),
            KWD(b"SCN"): (self.do_SCN, 0),
            KWD(b"TD"): (self.do_TD, 2),
            KWD(b"TJ"): (self.do_TJ, 1),
            KWD(b"TL"): (self.do_TL, 1),
            KWD(b"T*"): (self.do_T_a, 0),
            KWD(b"Tc"): (self.do_Tc, 1),
            KWD(b"Td"): (self.do_Td, 2),
            KWD(b"Tf"): (self.do_Tf, 2),
            KWD(b"Tj"): (self.do_Tj, 1),
            KWD(b"Tm"): (self.do_Tm, 6),
            KWD(b"Tr"): (self.do_Tr, 1),
            KWD(b"Ts"): (self.do_Ts, 1),
            KWD(b"Tw"): (self.do_Tw, 1),
            KWD(b"Tz"): (self.do_Tz, 1),
            KWD(b"W"): (self.do_W, 0),
            KWD(b"W*"): (self.do_W_a, 0),
            KWD(b"'"): (self.do__q, 1),
            KWD(b'"'): (self.do__w, 3),
            KWD(b"b"): (self.do_b, 0),
            KWD(b"b*"): (self.do_b_a, 0),
            KWD(b"c"): (self.do_c, 6),
            KWD(b"cm"): (self.do_cm, 6),
            KWD(b"cs"): (self.do_cs, 1),
            KWD(b"d"): (self.do_d, 2),
            KWD(b"f"): (self.do_f, 0),
            KWD(b"f*"): (self.do_f_a, 0),
            KWD(b"g"): (self.do_g, 1),
            KWD(b"gs"): (self.do_gs, 1),
            KWD(b"h"): (self.do_h, 0),
            KWD(b"i"): (self.do_i, 1),
            KWD(b"j"): (self.do_j, 1),
            KWD(b"k"): (self.do_k, 4),
            KWD(b"l"): (self.do_l, 2),
            KWD(b"m"): (self.do_m, 2),
            KWD(b"n"): (self.do_n, 0),
            KWD(b"q"): (self.do_q, 0),
            KWD(b"re"): (self.do_re, 4),
            KWD(b"rg"): (self.do_rg, 3),
            KWD(b"ri"): (self.do_ri, 1),
            KWD(b"s"): (self.do_s, 0),
            KWD(b"sc"): (self.do_sc, 0),
            KWD(b"scn"): (self.do_scn, 0),
            KWD(b"sh"): (self.do_sh, 1),
            KWD(b"v"): (self.do_v, 4),
            KWD(b"w"): (self.do_w, 1),
            KWD(b"y"): (self.do_y, 4),
        }
        self.filter_classes = filter_classes
        self.restrict_ops = restrict_ops
        self.page = page
        self.parent_key = (
            page.attrs.get("StructParents") if parent_key is None else parent_key
        )
        self.contents = contents
        self.ignore_colours = ignore_colours
        self.init_resources(page, page.resources if resources is None else resources)
        self.init_state(page.ctm if ctm is None else ctm, gstate)
        self.parser = ContentParser(contents, page.doc)
        self._prev_text: Union[TextObject, None] = None
        self._curpos = 0  # Relative to the stream so it always starts at zero
        # These are private for the moment while we think about the interface
        self._mcstart: Dict[int, Tuple[Union[int, None], int]] = {}
        self._mcend: Dict[int, Tuple[Union[int, None], int]] = {}

    def init_resources(self, page: "Page", resources: Dict) -> None:
        """Prepare the fonts and XObjects listed in the Resource attribute."""
        self.resources = resources
        self.fontmap: Mapping[str, Font] = {}
        self.xobjmap = {}
        self.csmap: Dict[str, ColorSpace] = copy(PREDEFINED_COLORSPACE)
        self.extgstatemap = {}
        if not self.resources:
            return

        for k, v in dict_value(self.resources).items():
            mapping = resolve1(v)
            if mapping is None:
                log.warning("Missing %s mapping", k)
                continue
            if k == "ProcSet":
                continue
            # PDF 2.0, sec 7.8.3, Table 34, ProcSet is an array,
            # everything else are dictionaries
            if not isinstance(mapping, dict):
                log.warning("%s mapping not a dict: %r", k, mapping)
                continue
            if k == "Font":
                self.fontmap = FontMapping(mapping, _deref_document(page.docref))
            elif k == "ColorSpace":
                for csid, spec in mapping.items():
                    colorspace = get_colorspace(resolve1(spec), csid)
                    if colorspace is not None:
                        self.csmap[csid] = colorspace
            elif k == "XObject":
                self.xobjmap = mapping
            elif k == "ExtGState":
                self.extgstatemap = mapping

    def init_state(self, ctm: Matrix, gstate: Union[GraphicState, None] = None) -> None:
        self.gstack: List[Tuple[Matrix, GraphicState, TextState]] = []
        self.ctm = ctm
        # Note the copy here, a new interpreter *always* creates a new
        # graphics state (including text state and CTM)
        self.graphicstate = GraphicState() if gstate is None else copy(gstate)
        # Mutable text state (just the matrices) - this is not
        # supposed to exist outside BT/ET pairs, but we will tolerate
        # it.  In PDF 2.0 it gets saved/restored on the stack by q/Q
        self.textstate: TextState = TextState()
        # Current path (FIXME: is this in the graphics state too?)
        self.curpath: List[PathSegment] = []
        # argstack: stack for command arguments.
        self.argstack: List[PDFObject] = []
        # mcstack: stack for marked content sections.
        self.mcstack: Tuple[MarkedContent, ...] = ()

    def pop(self, n: int) -> List[PDFObject]:
        x = self.argstack[-n:]
        del self.argstack[-n:]
        return x

    def __next__(self) -> ContentObject:
        while True:
            # Update glyph offset (used to be done with a coroutine)
            if self._prev_text is not None:
                self.textstate.glyph_offset = self._prev_text._get_next_glyph_offset()
                self._prev_text = None
            self._curpos, obj = next(self.parser)
            # These are handled inside the parser as they don't obey
            # the normal syntax rules (PDF 1.7 sec 8.9.7)
            if isinstance(obj, InlineImage):
                co = self.do_EI(obj)
                if co is not None:
                    return co
                continue
            if isinstance(obj, PSKeyword):
                co = self.operate(obj)
                if co is not None:
                    # Store a TextObject to update text state if we return
                    if isinstance(co, TextObject):
                        self._prev_text = co
                    if self.filter_classes is None or isinstance(
                        co, tuple(self.filter_classes)
                    ):
                        return co
                continue
            self.argstack.append(obj)

    def operate(self, obj: PSKeyword) -> Union[ContentObject, None]:
        if obj not in self._dispatch:
            return None
        method, nargs = self._dispatch[obj]
        if self.restrict_ops is not None and obj not in self.restrict_ops:
            # We don't care about the args at all!
            del self.argstack[-nargs:]
            return None
        if nargs == 0:
            return method()
        # Avoid self.pop() here in the fast path
        args = self.argstack[-nargs:]
        del self.argstack[-nargs:]
        if len(args) != nargs:
            log.warning(
                "Insufficient arguments (%d) for operator: %r",
                len(args),
                obj,
            )
            return None
        else:
            try:
                return method(*args)
            except TypeError as e:
                log.warning(
                    "Incorrect type of arguments(%r) for operator %r: %s",
                    args,
                    obj,
                    e,
                )
                return None

    def create(self, object_class, **kwargs) -> Union[ContentObject, None]:
        if self.filter_classes is not None and object_class not in self.filter_classes:
            return None
        return object_class(
            _pageref=self.page.pageref,
            _parentkey=self.parent_key,
            ctm=self.ctm,
            mcstack=self.mcstack,
            gstate=self.graphicstate,
            **kwargs,
        )

    def do_S(self) -> Union[ContentObject, None]:
        """Stroke path"""
        if not self.curpath:
            return None
        curpath = self.curpath
        self.curpath = []
        return self.create(
            PathObject,
            stroke=True,
            fill=False,
            evenodd=False,
            raw_segments=curpath,
        )

    def do_s(self) -> Union[ContentObject, None]:
        """Close and stroke path"""
        self.do_h()
        return self.do_S()

    def do_f(self) -> Union[ContentObject, None]:
        """Fill path using nonzero winding number rule"""
        if not self.curpath:
            return None
        curpath = self.curpath
        self.curpath = []
        return self.create(
            PathObject,
            stroke=False,
            fill=True,
            evenodd=False,
            raw_segments=curpath,
        )

    def do_F(self) -> Union[ContentObject, None]:
        """Fill path using nonzero winding number rule (obsolete)"""
        return self.do_f()

    def do_f_a(self) -> Union[ContentObject, None]:
        """Fill path using even-odd rule"""
        if not self.curpath:
            return None
        curpath = self.curpath
        self.curpath = []
        return self.create(
            PathObject,
            stroke=False,
            fill=True,
            evenodd=True,
            raw_segments=curpath,
        )

    def do_B(self) -> Union[ContentObject, None]:
        """Fill and stroke path using nonzero winding number rule"""
        if not self.curpath:
            return None
        curpath = self.curpath
        self.curpath = []
        return self.create(
            PathObject,
            stroke=True,
            fill=True,
            evenodd=False,
            raw_segments=curpath,
        )

    def do_B_a(self) -> Union[ContentObject, None]:
        """Fill and stroke path using even-odd rule"""
        if not self.curpath:
            return None
        curpath = self.curpath
        self.curpath = []
        return self.create(
            PathObject,
            stroke=True,
            fill=True,
            evenodd=True,
            raw_segments=curpath,
        )

    def do_b(self) -> Union[ContentObject, None]:
        """Close, fill, and stroke path using nonzero winding number rule"""
        self.do_h()
        return self.do_B()

    def do_b_a(self) -> Union[ContentObject, None]:
        """Close, fill, and stroke path using even-odd rule"""
        self.do_h()
        return self.do_B_a()

    def do_TJ(self, strings: PDFObject) -> Union[ContentObject, None]:
        """Show one or more text strings, allowing individual glyph
        positioning"""
        args: List[Union[bytes, float]] = []
        has_text = False
        for s in list_value(strings):
            if isinstance(s, (int, float)):
                args.append(s)
            elif isinstance(s, bytes):
                if s:
                    has_text = True
                args.append(s)
            else:
                log.warning(
                    "Ignoring non-string/number %r in text object %r", s, strings
                )
        obj = self.create(
            TextObject,
            line_matrix=self.textstate.line_matrix,
            _glyph_offset=self.textstate.glyph_offset,
            args=args,
        )
        if obj is not None:
            if has_text:
                return obj
            # Even without text, TJ can still update the glyph offset
            assert isinstance(obj, TextObject)
            self.textstate.glyph_offset = obj._get_next_glyph_offset()
        return None

    def do_Tj(self, s: PDFObject) -> Union[ContentObject, None]:
        """Show a text string"""
        return self.do_TJ([s])

    def do__q(self, s: PDFObject) -> Union[ContentObject, None]:
        """Move to next line and show text

        The ' (single quote) operator.
        """
        self.do_T_a()
        return self.do_TJ([s])

    def do__w(
        self, aw: PDFObject, ac: PDFObject, s: PDFObject
    ) -> Union[ContentObject, None]:
        """Set word and character spacing, move to next line, and show text

        The " (double quote) operator.
        """
        self.do_Tw(aw)
        self.do_Tc(ac)
        return self.do_TJ([s])

    def do_EI(self, obj: PDFObject) -> Union[ContentObject, None]:
        """End inline image object"""
        if isinstance(obj, InlineImage):
            # Inline images are not XObjects, have no xobjid
            return self.render_image(None, obj)
        else:
            log.warning("EI has unknown argument type: %r", obj)
            return None

    def do_Do(self, xobjid_arg: PDFObject) -> Union[ContentObject, None]:
        """Invoke named XObject"""
        xobjid = literal_name(xobjid_arg)
        try:
            xobj = stream_value(self.xobjmap[xobjid])
        except KeyError:
            log.debug("Undefined xobject id: %r", xobjid)
            return None
        except TypeError as e:
            raise TypeError(f"Empty or invalid xobject with id {xobjid!r}") from e
        subtype = xobj.get("Subtype")
        if subtype is LITERAL_FORM:
            # PDF Ref 1.7, # 4.9
            #
            # When the Do operator is applied to a form XObject,
            # it does the following tasks:
            #
            # 1. Saves the current graphics state, as if by invoking the q operator
            # ...
            # 5. Restores the saved graphics state, as if by invoking the Q operator
            #
            # The lazy interpretation of this is, obviously, that
            # we simply create an XObjectObject with a copy of the
            # current graphics state.  The copying is actually
            # done (lazily, of course) when construcitng a new
            # LazyInterpreter, not here.
            return XObjectObject.from_stream(
                stream=xobj,
                page=self.page,
                xobjid=xobjid,
                ctm=self.ctm,
                gstate=self.graphicstate,
                mcstack=self.mcstack,
            )
        elif subtype is LITERAL_IMAGE:
            return self.render_image(xobjid, xobj)
        else:
            log.debug("Unsupported XObject %r of type %r: %r", xobjid, subtype, xobj)
        return None

    def render_image(
        self, xobjid: Union[str, None], stream: ContentStream
    ) -> Union[ContentObject, None]:
        # Look up colorspace in resources first!
        cspec = resolve1(stream.get_any(("CS", "ColorSpace")))
        if isinstance(cspec, PSLiteral) and cspec.name in self.csmap:
            colorspace: Union[ColorSpace, None] = self.csmap[cspec.name]
        else:
            colorspace = get_colorspace(cspec)
        # Cache it in the stream object to avoid confusion
        if colorspace is not None:
            stream.colorspace = colorspace
        obj = self.create(
            ImageObject,
            stream=stream,
            xobjid=xobjid,
            srcsize=(stream.width, stream.height),
            # Ensure it is always a bool (whither mypy here?)
            imagemask=not not stream.get_any(("IM", "ImageMask")),
            bits=stream.bits,
            colorspace=colorspace,
        )
        # Override parent key if one is defined on the image specifically
        if obj is not None and "StructParent" in stream:
            obj._parentkey = int_value(stream["StructParent"])
        return obj

    def do_q(self) -> None:
        """Save graphics state"""
        self.gstack.append((self.ctm, copy(self.graphicstate), copy(self.textstate)))

    def do_Q(self) -> None:
        """Restore graphics state"""
        if self.gstack:
            self.ctm, self.graphicstate, self.textstate = self.gstack.pop()

    def do_cm(
        self,
        a1: PDFObject,
        b1: PDFObject,
        c1: PDFObject,
        d1: PDFObject,
        e1: PDFObject,
        f1: PDFObject,
    ) -> None:
        """Concatenate matrix to current transformation matrix"""
        cm = (
            num_value(a1),
            num_value(b1),
            num_value(c1),
            num_value(d1),
            num_value(e1),
            num_value(f1),
        )
        self.ctm = mult_matrix(cm, self.ctm)

    def do_w(self, linewidth: PDFObject) -> None:
        """Set line width"""
        self.graphicstate.linewidth = num_value(linewidth)

    def do_J(self, linecap: PDFObject) -> None:
        """Set line cap style"""
        self.graphicstate.linecap = int_value(linecap)

    def do_j(self, linejoin: PDFObject) -> None:
        """Set line join style"""
        self.graphicstate.linejoin = int_value(linejoin)

    def do_M(self, miterlimit: PDFObject) -> None:
        """Set miter limit"""
        self.graphicstate.miterlimit = num_value(miterlimit)

    def do_d(self, dash: PDFObject, phase: PDFObject) -> None:
        """Set line dash pattern"""
        ndash = tuple(num_value(x) for x in list_value(dash))
        self.graphicstate.dash = DashPattern(ndash, num_value(phase))

    def do_ri(self, intent: PDFObject) -> None:
        """Set color rendering intent"""
        if self.ignore_colours:
            return
        if isinstance(intent, PSLiteral):
            # Should possibly check that it is a valid intent
            self.graphicstate.intent = intent
        else:
            raise TypeError(f"Not a name: {intent!r}")

    def do_i(self, flatness: PDFObject) -> None:
        """Set flatness tolerance"""
        self.graphicstate.flatness = num_value(flatness)

    def do_gs(self, name: PDFObject) -> None:
        """Set parameters from graphics state parameter dictionary"""
        try:
            extgstate = dict_value(self.extgstatemap[literal_name(name)])
        except KeyError:
            log.warning("Undefined ExtGState: %r", name)
            return
        # PDF 2.0, sec 8.4.5, Table 57
        # Skipping Device-dependent graphics state parameters except
        # for flatness tolerance
        if "LW" in extgstate:
            self.do_w(extgstate["LW"])
        if "LC" in extgstate:
            self.do_J(extgstate["LC"])
        if "LJ" in extgstate:
            self.do_j(extgstate["LJ"])
        if "ML" in extgstate:
            self.do_M(extgstate["ML"])
        if "D" in extgstate:
            dash, phase = list_value(extgstate["D"])
            self.do_d(dash, phase)
        if "RI" in extgstate:
            self.do_ri(extgstate["RI"])
        if "Font" in extgstate:
            fontref, fontsize = list_value(extgstate["Font"])
            self.graphicstate.font = self.page.doc.get_font(fontref.objid, None)
            self.graphicstate.fontsize = num_value(fontsize)
        if "FL" in extgstate:
            self.do_i(extgstate["FL"])
        if "SA" in extgstate:
            self.graphicstate.stroke_adjustment = bool_value(extgstate["SA"])
        if "BM" in extgstate:
            bm = extgstate["BM"]
            if isinstance(bm, PSLiteral):
                self.graphicstate.blend_mode = bm
            else:
                bml: List[PSLiteral] = []
                for x in list_value(bm):
                    if isinstance(PSLiteral, x):
                        raise TypeError(f"Not a name: {x!r}")
                    bml.append(x)
                self.graphicstate.blend_mode = bml
        if "SMask" in extgstate:
            smask = extgstate["SMask"]
            if isinstance(smask, PSLiteral):
                self.graphicstate.smask = None
            else:
                self.graphicstate.smask = dict_value(smask)
        if "CA" in extgstate:
            self.graphicstate.salpha = num_value(extgstate["CA"])
        if "ca" in extgstate:
            self.graphicstate.nalpha = num_value(extgstate["ca"])
        if "AIS" in extgstate:
            self.graphicstate.alpha_source = bool_value(extgstate["AIS"])
        if "TK" in extgstate:
            self.graphicstate.knockout = bool_value(extgstate["TK"])
        if "UseBlackPtComp" in extgstate and not self.ignore_colours:
            black_pt_comp = extgstate["UseBlackPtComp"]
            assert isinstance(black_pt_comp, PSLiteral)
            self.graphicstate.black_pt_comp = black_pt_comp
        # Also ignored with ignore_colours, but we do not support
        # them: TR, TR2, HT, BG, BG2, UCR, UCR2

    def do_m(self, x: PDFObject, y: PDFObject) -> None:
        """Begin new subpath"""
        self.curpath.append(make_seg("m", (num_value(x), num_value(y))))

    def do_l(self, x: PDFObject, y: PDFObject) -> None:
        """Append straight line segment to path"""
        self.curpath.append(make_seg("l", (num_value(x), num_value(y))))

    def do_c(
        self,
        x1: PDFObject,
        y1: PDFObject,
        x2: PDFObject,
        y2: PDFObject,
        x3: PDFObject,
        y3: PDFObject,
    ) -> None:
        """Append curved segment to path (three control points)"""
        self.curpath.append(
            make_seg(
                "c",
                (num_value(x1), num_value(y1)),
                (num_value(x2), num_value(y2)),
                (num_value(x3), num_value(y3)),
            ),
        )

    def do_v(self, x2: PDFObject, y2: PDFObject, x3: PDFObject, y3: PDFObject) -> None:
        """Append curved segment to path (initial point replicated)"""
        self.curpath.append(
            make_seg(
                "v",
                (num_value(x2), num_value(y2)),
                (num_value(x3), num_value(y3)),
            )
        )

    def do_y(self, x1: PDFObject, y1: PDFObject, x3: PDFObject, y3: PDFObject) -> None:
        """Append curved segment to path (final point replicated)"""
        self.curpath.append(
            make_seg(
                "y",
                (num_value(x1), num_value(y1)),
                (num_value(x3), num_value(y3)),
            )
        )

    def do_h(self) -> None:
        """Close subpath"""
        self.curpath.append(make_seg("h"))

    def do_re(self, x: PDFObject, y: PDFObject, w: PDFObject, h: PDFObject) -> None:
        """Append rectangle to path"""
        x = num_value(x)
        y = num_value(y)
        w = num_value(w)
        h = num_value(h)
        self.curpath.append(make_seg("m", (x, y)))
        self.curpath.append(make_seg("l", (x + w, y)))
        self.curpath.append(make_seg("l", (x + w, y + h)))
        self.curpath.append(make_seg("l", (x, y + h)))
        self.curpath.append(make_seg("h"))

    def do_n(self) -> None:
        """End path without filling or stroking"""
        self.curpath = []

    def do_W(self) -> None:
        """Set clipping path using nonzero winding number rule"""

    def do_W_a(self) -> None:
        """Set clipping path using even-odd rule"""

    def do_CS(self, name: PDFObject) -> None:
        """Set color space for stroking operators

        Introduced in PDF 1.1
        """
        if self.ignore_colours:
            return
        try:
            self.graphicstate.scs = self.csmap[literal_name(name)]
        except KeyError:
            log.warning("Undefined ColorSpace: %r", name)

    def do_cs(self, name: PDFObject) -> None:
        """Set color space for nonstroking operators"""
        if self.ignore_colours:
            return
        try:
            self.graphicstate.ncs = self.csmap[literal_name(name)]
        except KeyError:
            log.warning("Undefined ColorSpace: %r", name)

    def do_G(self, gray: PDFObject) -> None:
        """Set gray level for stroking operators"""
        if self.ignore_colours:
            return
        self.graphicstate.scs = self.csmap["DeviceGray"]
        self.graphicstate.scolor = self.graphicstate.scs.make_color(gray)

    def do_g(self, gray: PDFObject) -> None:
        """Set gray level for nonstroking operators"""
        if self.ignore_colours:
            return
        self.graphicstate.ncs = self.csmap["DeviceGray"]
        self.graphicstate.ncolor = self.graphicstate.ncs.make_color(gray)

    def do_RG(self, r: PDFObject, g: PDFObject, b: PDFObject) -> None:
        """Set RGB color for stroking operators"""
        if self.ignore_colours:
            return
        self.graphicstate.scs = self.csmap["DeviceRGB"]
        self.graphicstate.scolor = self.graphicstate.scs.make_color(r, g, b)

    def do_rg(self, r: PDFObject, g: PDFObject, b: PDFObject) -> None:
        """Set RGB color for nonstroking operators"""
        if self.ignore_colours:
            return
        self.graphicstate.ncs = self.csmap["DeviceRGB"]
        self.graphicstate.ncolor = self.graphicstate.ncs.make_color(r, g, b)

    def do_K(self, c: PDFObject, m: PDFObject, y: PDFObject, k: PDFObject) -> None:
        """Set CMYK color for stroking operators"""
        if self.ignore_colours:
            return
        self.graphicstate.scs = self.csmap["DeviceCMYK"]
        self.graphicstate.scolor = self.graphicstate.scs.make_color(c, m, y, k)

    def do_k(self, c: PDFObject, m: PDFObject, y: PDFObject, k: PDFObject) -> None:
        """Set CMYK color for nonstroking operators"""
        if self.ignore_colours:
            return
        self.graphicstate.ncs = self.csmap["DeviceCMYK"]
        self.graphicstate.ncolor = self.graphicstate.ncs.make_color(c, m, y, k)

    def do_SCN(self) -> None:
        """Set color for stroking operators."""
        if self.ignore_colours:
            return
        assert self.graphicstate.scs is not None  # it is always not None now
        self.graphicstate.scolor = self.graphicstate.scs.make_color(
            *self.pop(self.graphicstate.scs.ncomponents)
        )

    def do_scn(self) -> None:
        """Set color for nonstroking operators"""
        if self.ignore_colours:
            return
        assert self.graphicstate.ncs is not None  # it is always not None now
        self.graphicstate.ncolor = self.graphicstate.ncs.make_color(
            *self.pop(self.graphicstate.ncs.ncomponents)
        )

    def do_SC(self) -> None:
        """Set color for stroking operators"""
        self.do_SCN()

    def do_sc(self) -> None:
        """Set color for nonstroking operators"""
        self.do_scn()

    def do_sh(self, name: Any) -> None:
        """Paint area defined by shading pattern"""
        if self.ignore_colours:
            return
        log.debug("sh operator currently unsupported")

    def do_BT(self) -> None:
        """Begin text object.

        Initializing the text matrix, Tm, and the text line matrix,
        Tlm, to the identity matrix. Text objects cannot be nested; a
        second BT cannot appear before an ET.  While Tm and Tlm are
        saved and restored with the graphics state, they do not
        persist outside a BT/ET pair.

        """
        self.textstate.reset()

    def do_ET(self) -> None:
        """End a text object"""

    def do_BX(self) -> None:
        """Begin compatibility section"""

    def do_EX(self) -> None:
        """End compatibility section"""

    def do_Tc(self, space: PDFObject) -> None:
        """Set character spacing.

        Character spacing is used by the Tj, TJ, and ' operators.

        :param space: a number expressed in unscaled text space units.
        """
        self.graphicstate.charspace = num_value(space)

    def do_Tw(self, space: PDFObject) -> None:
        """Set the word spacing.

        Word spacing is used by the Tj, TJ, and ' operators.

        :param space: a number expressed in unscaled text space units
        """
        self.graphicstate.wordspace = num_value(space)

    def do_Tz(self, scale: PDFObject) -> None:
        """Set the horizontal scaling.

        :param scale: is a number specifying the percentage of the normal width
        """
        self.graphicstate.scaling = num_value(scale)

    def do_TL(self, leading: PDFObject) -> None:
        """Set the text leading.

        Text leading is used only by the T*, ', and " operators.

        :param leading: a number expressed in unscaled text space units
        """
        self.graphicstate.leading = num_value(leading)

    def do_Tf(self, fontid: PDFObject, fontsize: PDFObject) -> None:
        """Set the text font

        :param fontid: the name of a font resource in the Font subdictionary
            of the current resource dictionary
        :param fontsize: size is a number representing a scale factor.
        """
        try:
            self.graphicstate.font = self.fontmap[literal_name(fontid)]
        except KeyError:
            log.warning("Undefined Font id: %r", fontid)
            doc = _deref_document(self.page.docref)
            self.graphicstate.font = doc.get_font()
        self.graphicstate.fontsize = num_value(fontsize)

    def do_Tr(self, render: PDFObject) -> None:
        """Set the text rendering mode"""
        self.graphicstate.render_mode = int_value(render)

    def do_Ts(self, rise: PDFObject) -> None:
        """Set the text rise

        :param rise: a number expressed in unscaled text space units
        """
        self.graphicstate.rise = num_value(rise)

    def do_Td(self, tx: PDFObject, ty: PDFObject) -> None:
        """Move to the start of the next line

        Offset from the start of the current line by (tx , ty).
        """
        try:
            tx = num_value(tx)
            ty = num_value(ty)
            (a, b, c, d, e, f) = self.textstate.line_matrix
            e_new = tx * a + ty * c + e
            f_new = tx * b + ty * d + f
            self.textstate.line_matrix = (a, b, c, d, e_new, f_new)
        except TypeError as e:
            raise TypeError(f"Invalid offset ({tx!r}, {ty!r})") from e
        self.textstate.glyph_offset = (0, 0)

    def do_TD(self, tx: PDFObject, ty: PDFObject) -> None:
        """Move to the start of the next line.

        offset from the start of the current line by (tx , ty). As a side effect, this
        operator sets the leading parameter in the text state.

        (PDF 1.7 Table 108) This operator shall have the same effect as this code:
            −ty TL
            tx ty Td
        """
        self.do_TL(-num_value(ty))
        self.do_Td(tx, ty)

    def do_Tm(
        self,
        a: PDFObject,
        b: PDFObject,
        c: PDFObject,
        d: PDFObject,
        e: PDFObject,
        f: PDFObject,
    ) -> None:
        """Set text matrix and text line matrix"""
        self.textstate.line_matrix = (
            num_value(a),
            num_value(b),
            num_value(c),
            num_value(d),
            num_value(e),
            num_value(f),
        )
        self.textstate.glyph_offset = (0, 0)

    def do_T_a(self) -> None:
        """Move to start of next text line"""
        (a, b, c, d, e, f) = self.textstate.line_matrix
        self.textstate.line_matrix = (
            a,
            b,
            c,
            d,
            -self.graphicstate.leading * c + e,
            -self.graphicstate.leading * d + f,
        )
        self.textstate.glyph_offset = (0, 0)

    def do_BI(self) -> None:
        """Begin inline image object"""

    def do_ID(self) -> None:
        """Begin inline image data"""

    def get_property(self, prop: PSLiteral) -> Union[Dict, None]:
        if "Properties" in self.resources:
            props = dict_value(self.resources["Properties"])
            return dict_value(props.get(prop.name))
        return None

    def do_MP(self, tag: PDFObject) -> Union[ContentObject, None]:
        """Define marked-content point"""
        return self.do_DP(tag, None)

    def do_DP(
        self, tag: PDFObject, props: PDFObject = None
    ) -> Union[ContentObject, None]:
        """Define marked-content point with property list"""
        # See above
        if isinstance(props, PSLiteral):
            props = self.get_property(props)
        rprops = {} if props is None else dict_value(props)
        return TagObject(
            _pageref=self.page.pageref,
            _parentkey=self.parent_key,
            ctm=self.ctm,
            mcstack=self.mcstack,
            gstate=self.graphicstate,
            _mcs=MarkedContent(mcid=None, tag=literal_name(tag), props=rprops),
        )

    def begin_tag(self, tag: PDFObject, props: Dict[str, PDFObject]) -> None:
        """Handle beginning of tag, setting current MCID if any."""
        assert isinstance(tag, PSLiteral)
        if "MCID" in props:
            mcid = int_value(props["MCID"])
            self._mcstart[mcid] = (self.parser.streamid, self._curpos)
        else:
            mcid = None
        self.mcstack = (
            *self.mcstack,
            MarkedContent(mcid=mcid, tag=tag.name, props=props),
        )

    def do_BMC(self, tag: PDFObject) -> None:
        """Begin marked-content sequence"""
        self.begin_tag(tag, {})

    def do_BDC(self, tag: PDFObject, props: PDFObject) -> None:
        """Begin marked-content sequence with property list"""
        # PDF 1.7 sec 14.6.2: If any of the values are indirect
        # references to objects outside the content stream, the
        # property list dictionary shall be defined as a named
        # resource in the Properties subdictionary of the current
        # resource dictionary (see 7.8.3, “Resource Dictionaries”) and
        # referenced by name as the properties operand of the DP or
        # BDC operat

        if not isinstance(tag, PSLiteral):
            log.warning("Tag %r is not a name object, ignoring", tag)
            return None
        if isinstance(props, PSLiteral):
            propdict = self.get_property(props)
            if propdict is None:
                log.warning("Missing property list in tag %r: %r", tag, props)
                propdict = {}
        else:
            propdict = dict_value(props)
        self.begin_tag(tag, propdict)

    def do_EMC(self) -> None:
        """End marked-content sequence"""
        if self.mcstack:
            # It's not a list so we can't pop it
            mcs = self.mcstack[-1]
            if mcs.mcid is not None:
                self._mcend[mcs.mcid] = (self.parser.streamid, self._curpos)
            self.mcstack = self.mcstack[:-1]


# This should be in playa.fontprogram, but circular imports...
class Type3Interpreter(LazyInterpreter):
    """Interpret Type3 font programs."""

    width: float = 1000
    bbox: Union[Rect, None]

    def __init__(
        self,
        page: "Page",
        contents: Iterable[PDFObject],
        resources: Union[Dict, None] = None,
        ctm: Union[Matrix, None] = None,
        gstate: Union[GraphicState, None] = None,
        parent_key: Union[int, None] = None,
        ignore_colours: bool = False,
    ) -> None:
        super().__init__(
            page, contents, resources, ctm, gstate, parent_key, ignore_colours
        )
        self._dispatch[KWD(b"d0")] = (self.do_d0, 2)
        self._dispatch[KWD(b"d1")] = (self.do_d1, 6)

    def do_d0(self, wx: float, wy: float) -> None:
        """Simple Type3 font metrics operator."""
        self.width = wx
        self.bbox = None

    def do_d1(
        self, wx: float, wy: float, llx: float, lly: float, urx: float, ury: float
    ) -> None:
        """More complete Type3 font metrics operator that ignores colours."""
        self.width = wx
        self.ignore_colours = True
        self.bbox = (llx, lly, urx, ury)



"""Schemas for various content objects.

This module contains schemas (as TypedDict) for content from various
PLAYA objects.

"""

import dataclasses
from typing import List, Tuple, Union

try:
    # We only absolutely need this when using Pydantic TypeAdapter
    from typing_extensions import TypedDict
except ImportError:
    from typing import TypedDict

from playa.color import Color as _Color
from playa.color import ColorSpace as _ColorSpace
from playa.data._asobj import asobj
from playa.data.metadata import Font
from playa.content import DashPattern as _DashPattern
from playa.content import GlyphObject as _GlyphObject
from playa.content import GraphicState as _GraphicState
from playa.content import ImageObject as _ImageObject
from playa.content import MarkedContent as _MarkedContent
from playa.content import PathObject as _PathObject
from playa.content import PathSegment as _PathSegment
from playa.content import TagObject as _TagObject
from playa.content import TextObject as _TextObject
from playa.pdftypes import resolve_all, MATRIX_IDENTITY, Matrix, Point, Rect


class Text(TypedDict, total=False):
    chars: str
    """Unicode string representation of text."""
    bbox: Union[Rect, None]
    """Bounding rectangle for all glyphs in text."""
    matrix: Matrix
    """Text rendering matrix.  Note that the effective font size
    and the origin can be derived from this."""
    gstate: "GraphicState"
    """Graphic state."""
    mcstack: List["Tag"]
    """Stack of enclosing marked content sections."""


class GraphicState(TypedDict, total=False):
    clipping_path: None  # TODO
    """The current clipping path (sec. 8.5.4), default `None`"""
    linewidth: float
    """Line width in user space units (sec. 8.4.3.2), default 1"""
    linecap: int
    """Line cap style (sec. 8.4.3.3), default 0 """
    linejoin: int
    """Line join style (sec. 8.4.3.4), default 0"""
    miterlimit: float
    """Maximum length of mitered line joins (sec. 8.4.3.5), default 10"""
    dash: "DashPattern"
    """Dash pattern for stroking (sec 8.4.3.6), default solid line `((), 0)`"""
    intent: str
    """Rendering intent (sec. 8.6.5.8), default `RelativeColorimetric`'"""
    stroke_adjustment: bool
    """A flag specifying whether to compensate for possible rasterization
    effects when stroking a path with a line width that is small relative to
    the pixel resolution of the output device (sec. 10.7.5), default `False`"""
    blend_mode: Union[str, List[str]]
    """The current blend mode that shall be used in the transparent imaging
    model (sec. 11.3.5), default `Normal`"""
    smask: Union[None, dict]
    """A soft-mask dictionary (sec. 11.6.5.1) or None, default `None`"""
    salpha: float
    """The constant shape or constant opacity value used for stroking
    operations (sec. 11.3.7.2 & 11.6.4.4), default 1"""
    nalpha: float
    """The constant shape or constant opacity value used for non-stroking
    operations, default 1"""
    alpha_source: bool
    """A flag specifying whether the current soft mask and alpha constant
    parameters shall be interpreted as shape values (true) or opacity values
    (false), default `False`"""
    black_pt_comp: str
    """The black point compensation algorithm that shall be used when
    converting CIE-based colours (sec. 8.6.5.9), default `Default`"""
    flatness: float
    """The precision with which curves shall be rendered on
    the output device (sec. 10.6.2), default 1"""
    scolor: "Color"
    """Colour used for stroking operations, default black `((0,), None)`"""
    scs: "ColorSpace"
    """Colour space used for stroking operations, default `DeviceGray`"""
    ncolor: "Color"
    """Colour used for non-stroking operations, default black `((0,) None)`"""
    ncs: "ColorSpace"
    """Colour space used for non-stroking operations, default `DeviceGray`"""
    font: Font
    """Descriptor of current font."""
    fontsize: float
    """Font size in unscaled text space units (**not** in points, can
    be scaled using a text or glyph's `matrix` to obtain user space
    units), default if not present is 1.0."""
    charspace: float
    """Character spacing in unscaled text space units, default if not present is 0."""
    wordspace: float
    """Word spacing in unscaled text space units, default if not present is 0."""
    scaling: float
    """Horizontal scaling factor multiplied by 100, default if not present is 100."""
    leading: float
    """Leading in unscaled text space units, default if not present is 0."""
    render_mode: int
    """Text rendering mode (PDF 1.7 Table 106), default if not present is 0."""
    rise: float
    """Text rise (for super and subscript) in unscaled text space
    units, default if not present is 0."""
    knockout: bool
    """The text knockout flag, shall determine the behaviour of overlapping
    glyphs within a text object in the transparent imaging model (sec. 9.3.8),
    default `True`"""


class DashPattern(TypedDict, total=False):
    dash: Tuple[float, ...]
    """Lengths of dashes and gaps in user space units."""
    phase: float
    """Starting position in the dash pattern, default 0."""


class Color(TypedDict, total=False):
    values: Tuple[float, ...]
    """Component values."""
    pattern: Union[str, None]
    """Pattern name in page resources, if any."""


class ColorSpace(TypedDict, total=False):
    name: str
    """Name of colour space."""
    ncomponents: int
    """Number of components."""
    spec: list
    """Specification."""


class MarkedContent(TypedDict, total=False):
    name: str
    """Marked content section tag name."""
    mcid: int
    """Marked content section ID."""
    props: dict
    """Marked content property dictionary (without MCID)."""


class Tag(TypedDict, total=False):
    name: str
    """Tag name."""
    mcid: int
    """Marked content section ID."""
    props: dict
    """Marked content property dictionary (without MCID)."""


class Image(TypedDict, total=False):
    ctm: Matrix
    """Coordinate transformation matrix, default is `(1, 0, 0, 1, 0, 0)`"""
    xobject_name: str
    """Name of XObject in page resources, if any."""
    bbox: Union[Rect, None]
    """Bounding rectangle for image."""
    srcsize: Tuple[int, int]
    """Size of source image in pixels."""
    bits: int
    """Number of bits per component, if required (default 1)."""
    imagemask: bool
    """True if the image is a mask."""
    stream: dict
    """Content stream dictionary."""
    colorspace: Union[ColorSpace, None]
    """Colour space for this image, if required."""


class Path(TypedDict, total=False):
    ctm: Matrix
    """Coordinate transformation matrix, default is `(1, 0, 0, 1, 0, 0)`"""
    stroke: bool
    """True if the outline of the path is stroked."""
    fill: bool
    """True if the path is filled."""
    evenodd: bool
    """True if the filling of complex paths uses the even-odd
    winding rule, False if the non-zero winding number rule is
    used (PDF 1.7 section 8.5.3.3)"""
    segments: List["PathSegment"]
    """Path segments."""
    gstate: "GraphicState"
    """Graphic state."""
    mcstack: List["Tag"]
    """Stack of enclosing marked content sections."""


class PathSegment(TypedDict, total=False):
    operator: str
    """Normalized path operator (PDF 1.7 section 8.5.2).  Note that "re"
    will be expanded into its constituent line segments."""
    points: List[Point]
    """Point or control points for path segment."""


class Glyph(TypedDict, total=False):
    text: str
    """Unicode string representation of glyph, if any."""
    cid: int
    """Character ID for glyph in font."""
    bbox: Union[Rect, None]
    """Bounding rectangle."""
    matrix: Matrix
    """Rendering matrix for glyph.  Note that the effective font size
    and the origin can be derived from this."""
    displacement: Point
    """Displacement to origin of next glyph."""
    gstate: "GraphicState"
    """Graphic state."""
    mcstack: List["Tag"]
    """Stack of enclosing marked content sections."""


@asobj.register
def asobj_color(obj: _Color) -> Color:
    color = Color(values=obj.values)
    if obj.pattern is not None:
        color["pattern"] = obj.pattern
    return color


@asobj.register
def asobj_colorspace(obj: _ColorSpace) -> ColorSpace:
    cs = ColorSpace(name=obj.name, ncomponents=obj.ncomponents)
    if obj.spec:
        cs["spec"] = asobj(resolve_all(obj.spec))
    return cs


@asobj.register
def asobj_dashpattern(obj: _DashPattern) -> DashPattern:
    dash = DashPattern(dash=obj.dash)
    if obj.phase != 0:
        dash["phase"] = obj.phase
    return dash


GRAPHICSTATE_DEFAULTS = {f.name: f.default for f in dataclasses.fields(_GraphicState)}


@asobj.register
def asobj_gstate(obj: _GraphicState) -> GraphicState:
    gstate = GraphicState()
    if obj.font is not None:
        gstate["font"] = asobj(obj.font)
        # fontsize is always defined with font
        gstate["fontsize"] = obj.fontsize
    for field in (
        "clipping_path",
        "linewidth",
        "linecap",
        "linejoin",
        "miterlimit",
        "dash",
        "intent",
        "stroke_adjustment",
        "blend_mode",
        "smask",
        "salpha",
        "nalpha",
        "alpha_source",
        "black_pt_comp",
        "flatness",
        "scolor",
        "scs",
        "ncolor",
        "ncs",
        "charspace",
        "wordspace",
        "leading",
        "render_mode",
        "rise",
        "knockout",
    ):
        val = getattr(obj, field)
        if val != GRAPHICSTATE_DEFAULTS[field]:
            gstate[field] = asobj(val)
    return gstate


@asobj.register
def asobj_mcs(obj: _MarkedContent) -> MarkedContent:
    props = {k: v for k, v in obj.props.items() if k != "MCID"}
    tag = MarkedContent(name=obj.tag)
    if obj.mcid is not None:
        tag["mcid"] = obj.mcid
    if props:
        tag["props"] = props
    return tag


@asobj.register
def asobj_text(obj: _TextObject) -> Text:
    text = Text(
        chars=obj.chars,
        bbox=obj.bbox,
        matrix=obj.matrix,
    )
    gstate = asobj(obj.gstate)
    if gstate:
        text["gstate"] = gstate
    mcstack = [asobj(mcs) for mcs in obj.mcstack]
    if mcstack:
        text["mcstack"] = mcstack
    return text


@asobj.register
def asobj_tag(obj: _TagObject) -> Tag:
    props = {k: v for k, v in obj.mcs.props.items() if k != "MCID"}
    tag = Tag(name=obj.mcs.tag)
    if obj.mcs.mcid is not None:
        tag["mcid"] = obj.mcs.mcid
    if props:
        tag["props"] = props
    return tag


@asobj.register
def asobj_image(obj: _ImageObject) -> Image:
    img = Image(srcsize=obj.srcsize, bbox=obj.bbox, stream=asobj(obj.stream))
    if obj.ctm is not MATRIX_IDENTITY:
        img["ctm"] = obj.ctm
    if obj.xobjid is not None:
        img["xobject_name"] = obj.xobjid
    if obj.bits != 1:
        img["bits"] = obj.bits
    if obj.imagemask:
        img["imagemask"] = True
    if obj.colorspace is not None:
        img["colorspace"] = asobj(obj.colorspace)
    return img


@asobj.register
def asobj_path(obj: _PathObject) -> Path:
    path = Path(segments=[asobj(seg) for seg in obj.segments])
    gstate = asobj(obj.gstate)
    if gstate:
        path["gstate"] = gstate
    if obj.ctm is not MATRIX_IDENTITY:
        path["ctm"] = obj.ctm
    mcstack = [asobj(mcs) for mcs in obj.mcstack]
    if mcstack:
        path["mcstack"] = mcstack
    for attr in "stroke", "fill", "evenodd":
        val = getattr(obj, attr, False)
        if val:
            path[attr] = val
    return path


@asobj.register
def asobj_path_segment(obj: _PathSegment) -> PathSegment:
    seg = PathSegment(operator=obj.operator)
    if obj.points:
        seg["points"] = list(obj.points)
    return seg


@asobj.register
def asobj_glyph(obj: _GlyphObject) -> Glyph:
    glyph = Glyph(
        cid=obj.cid,
        bbox=obj.bbox,
        matrix=obj.matrix,
        displacement=obj.displacement,
    )
    # But there is a default graphic state
    gstate = asobj(obj.gstate)
    if gstate:
        glyph["gstate"] = gstate
    if obj.text:
        glyph["text"] = obj.text
    mcstack = [asobj(mcs) for mcs in obj.mcstack]
    if mcstack:
        glyph["mcstack"] = mcstack
    return glyph



from typing import Dict, Iterator, List, Mapping, Tuple, Union, ItemsView

from playa.pdftypes import (
    PDFObject,
    ObjRef,
    dict_value,
    int_value,
    list_value,
    str_value,
)
from playa.utils import choplist


# TODO: NameTree and NumberTree are nearly identical and should be
# refactored to a single base class.


def walk_number_tree(
    tree: Dict[str, PDFObject], key: Union[int, None] = None
) -> Iterator[Tuple[int, PDFObject]]:
    stack: List[PDFObject] = [tree]
    while stack:
        item = dict_value(stack.pop())
        if key is not None and "Limits" in item:
            (k1, k2) = list_value(item["Limits"])
            if key < k1 or k2 < key:
                continue
        if "Nums" in item:
            k: Union[int, ObjRef]
            v: PDFObject
            for k, v in choplist(2, list_value(item["Nums"])):
                yield int_value(k), v
        if "Kids" in item:
            stack.extend(reversed(list_value(item["Kids"])))


class NumberTreeItemsView(ItemsView[int, PDFObject]):
    _mapping: "NumberTree"

    def __iter__(self) -> Iterator[Tuple[int, PDFObject]]:
        yield from walk_number_tree(self._mapping._obj)


class NumberTree(Mapping[int, PDFObject]):
    """A PDF number tree.

    See Section 7.9.7 of the PDF 1.7 Reference.

    Raises:
        TypeError: If initialized with a non-dictionary.
    """

    def __init__(self, obj: PDFObject):
        self._obj = dict_value(obj)

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __iter__(self) -> Iterator[int]:
        for idx, _ in walk_number_tree(self._obj):
            yield idx

    def __getitem__(self, num: int) -> PDFObject:
        for idx, val in walk_number_tree(self._obj, num):
            if idx == num:
                return val
        raise KeyError(f"Number {num} not in tree")

    def items(self) -> NumberTreeItemsView:
        return NumberTreeItemsView(self)


def walk_name_tree(
    tree: Dict[str, PDFObject], key: Union[bytes, None] = None
) -> Iterator[Tuple[bytes, PDFObject]]:
    stack: List[PDFObject] = [tree]
    while stack:
        item = dict_value(stack.pop())
        if key is not None and "Limits" in item:
            (k1, k2) = list_value(item["Limits"])
            if key < k1 or k2 < key:
                continue
        if "Names" in item:
            k: Union[bytes, ObjRef]
            v: PDFObject
            for k, v in choplist(2, list_value(item["Names"])):
                yield str_value(k), v
        if "Kids" in item:
            stack.extend(reversed(list_value(item["Kids"])))


class NameTreeItemsView(ItemsView[bytes, PDFObject]):
    _mapping: "NameTree"

    def __iter__(self) -> Iterator[Tuple[bytes, PDFObject]]:
        yield from walk_name_tree(self._mapping._obj)


class NameTree(Mapping[bytes, PDFObject]):
    """A PDF name tree.

    See Section 7.9.6 of the PDF 1.7 Reference.

    Raises:
        TypeError: If initialized with a non-dictionary.
    """

    def __init__(self, obj: PDFObject):
        self._obj = dict_value(obj)

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __iter__(self) -> Iterator[bytes]:
        for name, _ in walk_name_tree(self._obj):
            yield name

    def __getitem__(self, key: bytes) -> PDFObject:
        for name, val in walk_name_tree(self._obj, key):
            if name == key:
                return val
        raise KeyError("Name %r not in tree" % key)

    def items(self) -> NameTreeItemsView:
        return NameTreeItemsView(self)



"""
Basic classes for PDF document parsing.
"""

import io
import itertools
import logging
import mmap
import re
from concurrent.futures import Executor
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    ItemsView,
    Iterable,
    Iterator,
    List,
    Mapping,
    Sequence,
    Set,
    Tuple,
    Union,
    overload,
)

from playa.data_structures import NameTree, NumberTree
from playa.exceptions import (
    PDFEncryptionError,
    PDFSyntaxError,
)
from playa.font import CIDFont, Font, TrueTypeFont, Type1Font, Type3Font
from playa.outline import Destination, Tree as Outlines
from playa.page import (
    DeviceSpace,
    Page,
)
from playa.parser import (
    NOTKEYWORD,
    LIT,
    IndirectObject,
    IndirectObjectParser,
    Lexer,
    ObjectParser,
    ObjectStreamParser,
    PDFObject,
    PSLiteral,
    Token,
    literal_name,
)
from playa.pdftypes import (
    ContentStream,
    DecipherCallable,
    InlineImage,
    ObjRef,
    dict_value,
    int_value,
    list_value,
    resolve1,
    str_value,
    stream_value,
)
from playa.security import SECURITY_HANDLERS
from playa.structure import Tree
from playa.utils import (
    decode_text,
    format_int_alpha,
    format_int_roman,
)
from playa.worker import (
    PageRef,
    _deref_document,
    _deref_page,
    _ref_document,
    _set_document,
    in_worker,
)
from playa.xref import (
    XRef,
    XRefFallback,
    XRefStream,
    XRefTable,
    INDOBJR,
    XREFR,
)

log = logging.getLogger(__name__)


# Some predefined literals and keywords (these can be defined wherever
# they are used as they are interned to the same objects)
LITERAL_PDF = LIT("PDF")
LITERAL_TEXT = LIT("Text")
LITERAL_FONT = LIT("Font")
LITERAL_TYPE1 = LIT("Type1")
LITERAL_MMTYPE1 = LIT("MMType1")
LITERAL_TYPE0 = LIT("Type0")
LITERAL_TYPE3 = LIT("Type3")
LITERAL_TRUETYPE = LIT("TrueType")
LITERAL_OBJSTM = LIT("ObjStm")
LITERAL_XREF = LIT("XRef")
LITERAL_CATALOG = LIT("Catalog")
LITERAL_PAGE = LIT("Page")
LITERAL_PAGES = LIT("Pages")
INHERITABLE_PAGE_ATTRS = {"Resources", "MediaBox", "CropBox", "Rotate"}


def _find_header(buffer: Union[bytes, mmap.mmap]) -> Tuple[bytes, int]:
    start = buffer.find(b"%PDF-")
    if start == -1:
        log.warning("Could not find b'%PDF-' header, is this a PDF?")
        return b"", 0
    return buffer[start : start + 8], start


def _open_input(fp: Union[BinaryIO, bytes]) -> Tuple[str, int, Union[bytes, mmap.mmap]]:
    if isinstance(fp, bytes):
        buffer: Union[bytes, mmap.mmap] = fp
    else:
        try:
            buffer = mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ)
        except io.UnsupportedOperation:
            log.warning("mmap not supported on %r, reading document into memory", fp)
            buffer = fp.read()
        except ValueError:
            raise
    hdr, offset = _find_header(buffer)
    log.debug("Found header at %d: %r", offset, hdr)
    try:
        version = hdr[5:].decode("ascii")
    except UnicodeDecodeError:
        log.warning("Version number in header %r contains non-ASCII characters", hdr)
        version = "1.0"
    if not re.match(r"\d\.\d", version):
        log.warning("Version number in header %r is invalid", hdr)
        version = "1.0"
    return version, offset, buffer


class Document(Mapping[int, PDFObject]):
    """Representation of a PDF document.

    PDF documents, at a basic level, are collections of indirect
    objects with numeric IDs.  Since these IDs are sparse, and do not
    need to be ordered, this is best represented as a mapping of `int`
    to `PDFObject`.  The specification also provides for "generation
    numbers" which can be used to track successive revisions of the
    same object.  In practice very few PDFs actually do this, so the
    most recent generation of an object is accessible simply by its
    object ID.  To find other objects, use the `objects` property.

    Since PDF documents can be very large and complex, merely creating
    a `Document` does very little aside from verifying that the
    password is correct and getting a minimal amount of metadata.  In
    general, PLAYA will try to open just about anything as a PDF, so
    you should not expect the constructor to fail here if you give it
    nonsense (something else may fail later on).

    Some metadata, such as the structure tree and page tree, will be
    loaded lazily and cached.  We do not handle modification of PDFs.

    Args:
      fp: File-like object in binary mode, or a buffer with binary data.
          Files will be read using `mmap` if possible.  They do not need
          to be seekable, as if `mmap` fails the entire file will simply
          be read into memory (so a pipe or socket ought to work).
      password: Password for decryption, if needed.
      space: the device space to use for interpreting content ("screen"
          or "page")

    Raises:
      TypeError: if `fp` is a file opened in text mode (don't do that!)
      PDFEncryptionError: if the PDF has an unsupported encryption scheme
      PDFPasswordIncorrect: if the password is incorrect

    """

    trailer: Dict[str, PDFObject]
    info: Dict[str, PDFObject]
    buffer: Union[bytes, mmap.mmap]  # FIXME: abstract type for this?
    space: DeviceSpace
    encryption: Union[Tuple[Tuple[bytes, bytes], Dict], None] = None
    decipher: Union[DecipherCallable, None] = None

    _fp: Union[BinaryIO, None] = None
    _pages: Union["PageList", None] = None
    _pool: Union[Executor, None] = None
    _ncpus: int = 1
    _batchsize: int = 4
    _catalog: Union[Dict[str, PDFObject], None] = None
    _outline: Union["Outlines", None] = None
    _destinations: Union["Destinations", None] = None
    _structure: Union["Tree", None] = None
    _fontmap: Union[Mapping[str, Font], None] = None
    _parser: Union[IndirectObjectParser, None] = None
    _xrefs: Union[List[XRef], None] = None
    _trailer_pos = -1
    _startxref_pos = -1

    def __enter__(self) -> "Document":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def close(self) -> None:
        # If we were opened from a file then close it
        if self._fp:
            self._fp.close()
            self._fp = None
        # Shutdown process pool
        if self._pool:
            self._pool.shutdown()
            self._pool = None

    def __init__(
        self,
        fp: Union[BinaryIO, bytes],
        password: str = "",
        space: DeviceSpace = "screen",
        _boss_id: int = 0,
    ) -> None:
        # Get this out of the way, eh
        if isinstance(fp, io.TextIOBase):
            raise TypeError("fp is not a binary file")

        if _boss_id:
            # Set this **right away** because it is needed to get
            # indirect object references right.
            _set_document(self, _boss_id)
            assert in_worker()

        # Initialize mutable properties
        self.space = space
        self.info = {}
        self._cached_objs: Dict[int, PDFObject] = {}
        self._parsed_objs: Dict[int, Tuple[List[PDFObject], int]] = {}
        self._cached_fonts: Dict[int, Font] = {}
        self._cached_inline_images: Dict[
            Tuple[int, int], Tuple[int, Union[InlineImage, None]]
        ] = {}
        self._pdf_version, self._offset, self.buffer = _open_input(fp)
        # These are always True unless "encryption" (lol) is present
        self.is_printable = self.is_modifiable = self.is_extractable = True
        # We are Lazy, only find and read the trailer.
        self.trailer = self._read_trailer()
        # If there is encryption, then we need to read xref tables.
        # Otherwise we will defer this to the first object lookup.
        if "Encrypt" in self.trailer:
            self._xrefs = self._read_xrefs()
            try:
                ids = list_value(self.trailer["ID"])
                id_value = (bytes(ids[0]), bytes(ids[1]))
            except (KeyError, TypeError):
                # Some documents may not have a /ID, use two empty
                # byte strings instead. Solves
                # https://github.com/pdfminer/pdfminer.six/issues/594
                id_value = (b"", b"")
            encrypt = dict_value(self.trailer["Encrypt"])
            self.encryption = (id_value, encrypt)
            self._initialize_password(password)

    def _read_trailer(self) -> Dict[str, Any]:
        # To read the trailer, we must first find the trailer, which
        # is supposed to be at the end of the file, immediately before
        # the "startxref" keyword and after a "trailer" keyword.  This,
        # like so many other things in the PDF standard, is a cruel
        # lie, because:
        #
        # 1. If the file only contains cross-reference streams, there
        #    is no "trailer" keyword, and the trailer is the stream
        #    dictionary (which could be anywhere in the file)
        # 2. If the file is a linearized PDF, there *is* a trailer at
        #    the end of the file, but it's a bogus one that only
        #    contains /Size. The real trailer is after the main xref
        #    table (or stream) pointed to by the "startxref" value.
        # 3. Of course nobody understood these rules and so you might
        #    find the trailer in various other places.  Also, the
        #    "startxref" value is probably wrong.
        end = len(self.buffer)
        indobj = -1
        for pos in range(len(self.buffer) - 1, -2, -1):
            if pos == -1 or self.buffer[pos] in NOTKEYWORD:
                token = self.buffer[pos + 1 : end]
                if token == b"startxref":
                    try:
                        _, val = next(ObjectParser(self.buffer, pos=end))
                    except StopIteration:
                        continue
                    self._startxref_pos = int_value(val)
                    self._startxref_pos += self._offset
                    # If this is an xref stream, then its dictionary
                    # is the trailer.
                    if m := INDOBJR.match(self.buffer, self._startxref_pos):
                        self._trailer_pos = m.end(0)
                        break
                    # If this is a normal xref table, then look for a
                    # trailer after it, which will be the correct one
                    # to use (because linearization)
                    if m := XREFR.match(self.buffer, self._startxref_pos):
                        self._trailer_pos = self.buffer.find(
                            b"trailer", self._startxref_pos
                        )
                        if self._trailer_pos != -1:
                            self._trailer_pos += 7
                            break
                if token == b"trailer":
                    # We continued to scan backwards and found a trailer
                    self._trailer_pos = end
                if token == b"obj":
                    # We continued to scan backwards and found an
                    # indirect object, which may or may not be the
                    # cross-reference stream.
                    if self._trailer_pos == -1:
                        self._trailer_pos = end
                    indobj = 0
                if token == b"xref":
                    # We continued to scan backwards and found an xref table
                    self._startxref_pos = pos + 1
                    break
                if indobj != -1 and ord("0") <= token[0] <= ord("9"):
                    # We are in the abovementioned indirect object
                    indobj += 1
                    if indobj == 2:
                        self._startxref_pos = pos + 1
                        break
                end = pos
        self._trailer_pos, trailer = next(
            ObjectParser(self.buffer, pos=self._trailer_pos, doc=self)
        )
        if not isinstance(trailer, dict):
            raise PDFSyntaxError(f"Trailer is not a dict: {trailer!r}")
        if indobj == 2 and trailer.get("Type") != LITERAL_XREF:
            # Either it's just the trailer (no problem) and there's no
            # xref table, or it's some other random indirect object.
            self._startxref_pos = -1
        return trailer

    def _initialize_password(self, password: str = "") -> None:
        """Initialize the decryption handler with a given password, if any.

        Internal function, requires the Encrypt dictionary to have
        been read from the trailer into self.encryption.
        """
        assert self.encryption is not None
        (docid, param) = self.encryption
        if literal_name(param.get("Filter")) != "Standard":
            raise PDFEncryptionError("Unknown filter: param=%r" % param)
        v = int_value(param.get("V", 0))
        # 3 (PDF 1.4) An unpublished algorithm that permits encryption
        # key lengths ranging from 40 to 128 bits. This value shall
        # not appear in a conforming PDF file.
        if v == 3:
            raise PDFEncryptionError("Unpublished algorithm 3 not supported")
        factory = SECURITY_HANDLERS.get(v)
        # 0 An algorithm that is undocumented. This value shall not be used.
        if factory is None:
            raise PDFEncryptionError("Unknown algorithm: param=%r" % param)
        handler = factory(docid, param, password)
        self.decipher = handler.decrypt
        self.is_printable = handler.is_printable
        self.is_modifiable = handler.is_modifiable
        self.is_extractable = handler.is_extractable
        # Ensure that no extra data leaks into encrypted streams
        self.parser.strict = True
        self.parser.decipher = self.decipher

    @property
    def parser(self) -> IndirectObjectParser:
        if self._parser is not None:
            return self._parser
        self._parser = IndirectObjectParser(self.buffer, doc=self)
        self._parser.seek(self._offset)
        return self._parser

    @property
    def xrefs(self) -> List[XRef]:
        if self._xrefs is not None:
            return self._xrefs
        self._xrefs = self._read_xrefs()
        size = sum(len(x) for x in self._xrefs)
        self.trailer["Size"] = size
        log.debug("Updated /Size in trailer to %d", size)
        return self._xrefs

    def _read_xrefs(self) -> List[XRef]:
        if self._startxref_pos == -1:
            log.warning("startxref was not found, falling back to object parser")
            return [XRefFallback(self)]
        self._xrefpos: Set[int] = set()
        xrefs: List[XRef] = []
        try:
            self._read_xrefs_into(self._startxref_pos, xrefs)
            return xrefs
        except (ValueError, IndexError, StopIteration, PDFSyntaxError) as e:
            log.warning("xref parsing failed, falling back to object parser: %s", e)
            return [XRefFallback(self)]

    def _read_xrefs_into(
        self,
        start: int,
        xrefs: List[XRef],
    ) -> None:
        """Reads XRefs from the given location."""
        if start in self._xrefpos:
            log.warning("Detected circular xref chain at %d", start)
            return
        # Look for an XRefStream first, then an XRefTable
        if INDOBJR.match(self.buffer, start):
            log.debug("Reading xref stream at %d", start)
            # XRefStream: PDF-1.5
            xref: XRef = XRefStream(self, pos=start, offset=self._offset)
        elif m := XREFR.match(self.buffer, start):
            log.debug("Reading xref table at %d", m.start(1))
            xref = XRefTable(self, pos=m.start(1), offset=self._offset)
        else:
            # Well, maybe it's an XRef table without "xref" (but
            # probably not)
            xref = XRefTable(self, pos=start, offset=self._offset)
        self._xrefpos.add(start)
        xrefs.append(xref)
        trailer = xref.trailer
        # For hybrid-reference files, an additional set of xrefs as a
        # stream.
        if "XRefStm" in trailer:
            pos = int_value(trailer["XRefStm"])
            self._read_xrefs_into(pos + self._offset, xrefs)
        # Recurse into any previous xref tables or streams
        if "Prev" in trailer:
            # find previous xref
            pos = int_value(trailer["Prev"])
            self._read_xrefs_into(pos + self._offset, xrefs)

    @property
    def catalog(self) -> Dict[str, Any]:
        if self._catalog is not None:
            return self._catalog
        self._catalog = {}
        if "Root" in self.trailer:
            # Every PDF file must have exactly one /Root dictionary.
            try:
                self._catalog = dict_value(self.trailer["Root"])
            except TypeError:
                log.warning("Root is a broken reference (incorrect xref table?)")
        else:
            log.warning("No /Root object! - Is this really a PDF?")
        if "Type" in self._catalog and self._catalog["Type"] is not LITERAL_CATALOG:
            log.warning(f"Catalog doesn't seem to be a catalog: {self._catalog!r}")
        return self._catalog

    @property
    def is_tagged(self) -> bool:
        markinfo = resolve1(self.catalog.get("MarkInfo"))
        if isinstance(markinfo, dict):
            return not not markinfo.get("Marked")
        return False

    @property
    def pdf_version(self) -> str:
        if "Version" in self.catalog:
            log.debug(
                "Using PDF version %r from catalog instead of %r from header",
                self.catalog["Version"],
                self._pdf_version,
            )
            return literal_name(self.catalog["Version"])
        return self._pdf_version

    def __len__(self) -> int:
        """Return the number of indirect objects in this PDF.

        Danger: This number is unreliable and ephemeral.
            In a conforming PDF, the number of objects is declared by
            the `/Size` key in the document trailer, but conforming
            PDFs do not exist in the real world.  Upon opening a PDF,
            the trailer value will be returned here, but once the
            cross-reference tables have been loaded, a *different*
            number will be returned, and in the case where the
            cross-reference tables are invalid and must be
            regenerated, this value will be updated yet again.  This
            is the price of laziness (not to be confused with the
            wages of sin).
        """
        size = self.trailer.get("Size", None)
        if isinstance(size, int):
            return size
        size = sum(len(x) for x in self.xrefs)
        self.trailer["Size"] = size
        log.debug("Updated /Size in trailer to %d", size)
        return size

    def __iter__(self) -> Iterator[int]:
        """Iterate over object IDs"""
        return itertools.chain.from_iterable(self.xrefs)

    @property
    def objects(self) -> Iterator[IndirectObject]:
        """Iterate over all indirect objects (including, then expanding object
        streams)"""
        for _, obj in IndirectObjectParser(
            self.buffer, self, pos=self._offset, strict=self.parser.strict
        ):
            yield obj
            if (
                isinstance(obj.obj, ContentStream)
                and obj.obj.get("Type") is LITERAL_OBJSTM
            ):
                parser = ObjectStreamParser(obj.obj, self)
                for _, sobj in parser:
                    yield sobj

    @property
    def tokens(self) -> Iterator[Token]:
        """Iterate over tokens."""
        return (tok for pos, tok in Lexer(self.buffer))

    @property
    def structure(self) -> Union[Tree, None]:
        """Logical structure of this document, if any.

        In the case where no logical structure tree exists, this will
        be `None`.  Otherwise you may iterate over it, search it, etc.

        We do this instead of simply returning an empty structure tree
        because the vast majority of PDFs have no logical structure.
        Also, because the structure is a lazy object (the type
        signature here may change to `Iterable[Element]` at some
        point) there is no way to know if it's empty without iterating
        over it.

        """
        if self._structure is not None:
            return self._structure
        try:
            self._structure = Tree(self)
        except (TypeError, KeyError):
            self._structure = None
        return self._structure

    def _getobj_objstm(
        self, stream: ContentStream, index: int, objid: int
    ) -> PDFObject:
        if stream.objid in self._parsed_objs:
            (objs, n) = self._parsed_objs[stream.objid]
        else:
            (objs, n) = self._get_objects(stream)
            assert stream.objid is not None
            self._parsed_objs[stream.objid] = (objs, n)
        i = n * 2 + index
        try:
            obj = objs[i]
        except IndexError as e:
            raise PDFSyntaxError(
                "index %d + %d too big for stream of %d objects"
                % (n * 2, index, len(objs))
            ) from e
        return obj

    def _get_objects(self, stream: ContentStream) -> Tuple[List[PDFObject], int]:
        if stream.get("Type") is not LITERAL_OBJSTM:
            log.warning("Content stream Type is not /ObjStm: %r" % stream)
        try:
            n = int_value(stream["N"])
        except KeyError:
            log.warning("N is not defined in content stream: %r" % stream)
            n = 0
        except TypeError:
            log.warning("N is invalid in content stream: %r" % stream)
            n = 0
        parser = ObjectParser(stream.buffer, self)
        objs: List[PDFObject] = [obj for _, obj in parser]
        return (objs, n)

    def _getobj_parse(self, pos: int, objid: int) -> PDFObject:
        self.parser.seek(pos)
        try:
            m = INDOBJR.match(self.buffer, pos)
            if m is None:
                raise PDFSyntaxError(
                    f"Not an indirect object at position {pos}: "
                    f"{self.buffer[pos : pos + 8]!r}"
                )
            _, obj = next(self.parser)
            if obj.objid != objid:
                raise PDFSyntaxError(f"objid mismatch: {obj.objid!r}={objid!r}")
        except (ValueError, IndexError, PDFSyntaxError) as e:
            raise PDFSyntaxError(
                "Indirect object %d not found at position %d"
                % (
                    objid,
                    pos,
                )
            ) from e
        if obj.objid != objid:
            raise PDFSyntaxError(f"objid mismatch: {obj.objid!r}={objid!r}")
        return obj.obj

    def __getitem__(self, objid: int) -> PDFObject:
        """Get an indirect object from the PDF.

        Note that the behaviour in the case of a non-existent object
        (raising `KeyError`), while Pythonic, is not PDFic, as PDF
        1.7 sec 7.3.10 states:

        > An indirect reference to an undefined object shall not be
        considered an error by a conforming reader; it shall be
        treated as a reference to the null object.

        Raises:
          ValueError: if Document cannot be initialized
          KeyError: if objid does not exist in PDF

        """
        if objid == 0:
            raise KeyError("PDF object id cannot be 0.")

        if objid in self._cached_objs:
            if self._cached_objs[objid] is None:
                raise KeyError(f"Object with ID {objid} not found")
            return self._cached_objs[objid]
        obj = None

        for xref in self.xrefs:
            try:
                (strmid, index, genno) = xref[objid]
            except KeyError:
                continue
            try:
                if strmid is not None:
                    stream = stream_value(self[strmid])
                    obj = self._getobj_objstm(stream, index, objid)
                else:
                    try:
                        obj = self._getobj_parse(index, objid)
                    except PDFSyntaxError as e:
                        log.warning(
                            "Indirect object %d not found at position %d: %r",
                            objid,
                            index,
                            e,
                        )
                        # xref tables are clearly borked, so
                        # rebuild them and try again
                        log.warning("Rebuilding xref table from object parser")
                        fallback = XRefFallback(self)
                        self.trailer["Size"] = len(fallback)
                        log.debug(
                            "Updated /Size in trailer to %d", self.trailer["Size"]
                        )
                        self._xrefs = [fallback]
                        try:
                            (strmid, index, genno) = self._xrefs[0][objid]
                            obj = self._getobj_parse(index, objid)
                        except (KeyError, PDFSyntaxError) as e:
                            log.warning(
                                "Indirect object %d STILL not found at position %d: %r",
                                objid,
                                index,
                                e,
                            )
                break
            # FIXME: We might not actually want to catch these...
            except StopIteration:
                log.debug("EOF when searching for object %d", objid)
                continue
            except PDFSyntaxError as e:
                log.debug("Syntax error when searching for object %d: %s", objid, e)
                continue

        # Store it anyway as None if we can't find it to avoid costly searching
        self._cached_objs[objid] = obj
        if obj is None:
            raise KeyError(f"Object with ID {objid} not found")
        return self._cached_objs[objid]

    def get_font(
        self, objid: int = 0, spec: Union[Dict[str, PDFObject], None] = None
    ) -> Font:
        if objid and objid in self._cached_fonts:
            return self._cached_fonts[objid]
        if spec is None:
            return Font({}, {})
        # Create a Font object, hopefully
        font: Union[Font, None] = None
        if spec.get("Type") is not LITERAL_FONT:
            log.warning("Font Type is not /Font: %r", spec)
        subtype = spec.get("Subtype")
        if subtype in (LITERAL_TYPE1, LITERAL_MMTYPE1):
            font = Type1Font(spec)
        elif subtype is LITERAL_TRUETYPE:
            font = TrueTypeFont(spec)
        elif subtype == LITERAL_TYPE3:
            font = Type3Font(spec)
        elif subtype == LITERAL_TYPE0:
            if "DescendantFonts" not in spec:
                log.warning("Type0 font has no DescendantFonts: %r", spec)
            else:
                dfonts = list_value(spec["DescendantFonts"])
                if len(dfonts) != 1:
                    log.debug(
                        "Type 0 font should have 1 descendant, has more: %r", dfonts
                    )
                subspec = resolve1(dfonts[0])
                if not isinstance(subspec, dict):
                    log.warning("Invalid descendant font: %r", subspec)
                else:
                    subspec = subspec.copy()
                    # Merge the root and descendant font dictionaries
                    for k in ("Encoding", "ToUnicode"):
                        if k in spec:
                            subspec[k] = resolve1(spec[k])
                    font = CIDFont(subspec)
        else:
            log.warning("Unknown Subtype in font: %r" % spec)
        if font is None:
            # We need a dummy font object to be able to do *something*
            # (even if it's the wrong thing) with text objects.
            font = Font({}, {})
        if objid:
            self._cached_fonts[objid] = font
        return font

    @property
    def fonts(self) -> Mapping[str, Font]:
        """Get the mapping of font names to fonts for this document.

        Note that this can be quite slow the first time it's accessed
        as it must scan every single page in the document.

        Note: Font names may collide.
            Font names are generally understood to be globally unique
            <del>in the neighbourhood</del> in the document, but there's no
            guarantee that this is the case.  In keeping with the
            "incremental update" philosophy dear to PDF, you get the
            last font defined with a given name.
        """
        if self._fontmap is not None:
            return self._fontmap
        self._fontmap: Mapping[str, Font] = FontMapping(self)
        return self._fontmap

    @property
    def outline(self) -> Union[Outlines, None]:
        """Document outline, if any."""
        if "Outlines" not in self.catalog:
            return None
        if self._outline is not None:
            return self._outline
        try:
            self._outline = Outlines(self)
        except TypeError:
            log.warning(
                "Invalid Outlines entry in catalog: %r", self.catalog["Outlines"]
            )
            return None
        return self._outline

    @property
    def page_labels(self) -> Union[Iterator[str], None]:
        """Iterate over page label strings for the PDF document.

        If the document includes page labels, this generates strings,
        one per page, otherwise it is None.

        Warning: Unbounded iterator
            This iterator is unbounded, because the page label tree
            has no relation to the actual page tree, so it is
            recommended to use `pages` instead.
        """
        if self.catalog is None:
            return None
        label_tree_obj = self.catalog.get("PageLabels")
        if label_tree_obj is None:
            return None
        label_tree = NumberTree(label_tree_obj)
        return _iter_labels(label_tree)

    @property
    def pages(self) -> "PageList":
        """Pages of the document as an iterable/addressable `PageList` object."""
        if self._pages is None:
            self._pages = PageList(self)
        return self._pages

    @property
    def names(self) -> Dict[str, Any]:
        """PDF name dictionary (PDF 1.7 sec 7.7.4).

        Raises:
          KeyError: if nonexistent.
        """
        return dict_value(self.catalog["Names"])

    @property
    def destinations(self) -> "Destinations":
        """Named destinations as an iterable/addressable `Destinations` object."""
        if self._destinations is None:
            self._destinations = Destinations(self)
        return self._destinations


class FontMapping(Mapping[str, Font]):
    """Lazy mapping of font names to fonts in a Document."""

    def __init__(self, doc: Document) -> None:
        self._doc = doc
        self._fontmap: Dict[str, Union[Font, None]] = {}

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __iter__(self) -> Iterator[str]:
        for name, _ in self._iteritems():
            yield name

    def _iteritems(self) -> Iterator[Tuple[str, Font]]:
        unique_fontnames: Set[str] = set()
        for page in reversed(self._doc.pages):
            # Cache a whole page of fonts at a time
            page_fonts: List[Font] = list(page.fonts.values())
            fontnames: List[str] = []
            for font in reversed(page_fonts):
                if font.fontname not in self._fontmap:
                    self._fontmap[font.fontname] = font
                if font.fontname not in unique_fontnames:
                    fontnames.append(font.fontname)
                    unique_fontnames.add(font.fontname)
            # Now iterate over the name and (uniquified) font
            for name in fontnames:
                # STFU, mypy
                font_and_not_none = self._fontmap[name]
                assert font_and_not_none is not None
                yield name, font_and_not_none

    def __getitem__(self, fontname: str) -> Font:
        if fontname not in self._fontmap:
            # Yes, this is (worst-case) quadratic, but it's also Lazy.
            for name, font in self._iteritems():
                if name == fontname:
                    return font
            # We did not find it, so store None to avoid future finding
            self._fontmap[fontname] = None
        font_or_none = self._fontmap[fontname]
        if font_or_none is None:
            raise KeyError(f"Font {fontname} not found in document!")
        return font_or_none  # It cannot be None!!!


def call_page(func: Callable[[Page], Any], pageref: PageRef) -> Any:
    """Call a function on a page in a worker process."""
    return func(_deref_page(pageref))


class PageList(Sequence[Page]):
    """List of pages indexable by 0-based index or string label.

    Note that iteration over this object is lazy, but indexing is not.

    Attributes:
        have_labels: If pages have explicit labels in the PDF.
    """

    have_labels: bool
    _pages: Union[List[Page], None] = None
    _by_label: Union[Dict[str, Page], None] = None
    _by_objid: Union[Dict[int, Page], None] = None

    def __init__(
        self,
        doc: Union[Document, None] = None,
        pages: Union[Iterable[Page], None] = None,
    ) -> None:
        if doc is not None:
            self.docref = _ref_document(doc)
            self.have_labels = doc.page_labels is not None
        if pages is not None:
            self._pages = list(pages)
            self._by_label = {
                page.label: page for page in pages if page.label is not None
            }
            self._by_objid = {page.pageid: page for page in pages}

    def _eager_load(self) -> None:
        # Not for public consumption
        if self._pages is None:
            self._pages, self._by_label, self._by_objid = self._init_pages()

    def _init_pages(self) -> Tuple[List[Page], Dict[str, Page], Dict[int, Page]]:
        pages: List[Page] = []
        by_label: Dict[str, Page] = {}
        by_objid: Dict[int, Page] = {}
        doc = self.doc
        for page_idx, ((objid, properties), label) in enumerate(
            zip(self.page_objects, self.page_labels)
        ):
            page = Page(doc, objid, properties, label, page_idx, doc.space)
            pages.append(page)
            by_objid[objid] = page
            if label is not None:
                if label in by_label:
                    log.info("Duplicate page label %s at index %d", label, page_idx)
                else:
                    by_label[label] = page
        return pages, by_label, by_objid

    @property
    def page_labels(self) -> Iterator[str]:
        """Generate page labels from document catalog or make them up."""
        doc = self.doc
        if doc.page_labels is not None:
            yield from doc.page_labels
        else:
            yield from (str(idx) for idx in itertools.count(1))

    @property
    def page_objects(self) -> Iterator[Tuple[int, Dict[str, PDFObject]]]:
        """Iterate over the page tree yielding (objid, dictionary) tuples."""
        doc = self.doc
        # In the case of no page tree then make shit up based on the objects
        if "Pages" not in doc.catalog:
            for object_id, obj in self.doc.items():
                if isinstance(obj, dict) and obj.get("Type") is LITERAL_PAGE:
                    yield object_id, obj
            return
        stack = [(doc.catalog["Pages"], doc.catalog)]
        visited = set()
        while stack:
            (obj, parent) = stack.pop()
            if isinstance(obj, ObjRef):
                # The PDF specification *requires* both the Pages
                # element of the catalog and the entries in Kids in
                # the page tree to be indirect references.
                object_id = int(obj.objid)
            elif isinstance(obj, int):
                # Should not happen in a valid PDF, but probably does?
                log.warning("Page tree contains bare integer: %r in %r", obj, parent)
                object_id = obj
            elif obj is None:
                log.warning("Page tree contains null")
                object_id = -1
            else:
                log.warning("Page tree contains unknown object: %r", obj)
                object_id = -1
            try:
                page_object = dict_value(doc[object_id])
            except (KeyError, TypeError) as e:
                if object_id != -1:
                    log.warning("Missing or invalid page object %r: %s", object_id, e)
                # Create an empty page to match what pdfium does
                page_object = {
                    "Type": LIT("Page"),
                    "Resources": {},
                    "MediaBox": [0, 0, 612, 792],
                }

            # Avoid recursion errors by keeping track of visited nodes
            # (again, this should never actually happen in a valid PDF)
            if object_id in visited:
                log.warning("Circular reference %r in page tree", object_id)
                continue
            if object_id != -1:
                visited.add(object_id)

            # Propagate inheritable attributes
            object_properties = page_object.copy()
            for k, v in parent.items():
                if k in INHERITABLE_PAGE_ATTRS and k not in object_properties:
                    object_properties[k] = v

            # Recurse, depth-first
            object_type = object_properties.get("Type")
            if object_type is None:
                log.warning("Page has no Type, trying type: %r", object_properties)
                object_type = object_properties.get("type")
            if object_type is LITERAL_PAGES and "Kids" in object_properties:
                for child in reversed(list_value(object_properties["Kids"])):
                    stack.append((child, object_properties))
            elif object_type is LITERAL_PAGE:
                yield object_id, object_properties

    @property
    def doc(self) -> "Document":
        """Get associated document if it exists."""
        return _deref_document(self.docref)

    def __len__(self) -> int:
        if self._pages is not None:
            return len(self._pages)
        return sum(1 for _ in self.page_objects)

    def __iter__(self) -> Iterator[Page]:
        if self._pages is not None:
            yield from self._pages
            return
        doc = self.doc
        for page_idx, ((objid, properties), label) in enumerate(
            zip(self.page_objects, self.page_labels)
        ):
            yield Page(doc, objid, properties, label, page_idx, doc.space)

    @overload
    def __getitem__(self, key: int) -> Page: ...

    @overload
    def __getitem__(self, key: str) -> Page: ...

    @overload
    def __getitem__(self, key: slice) -> "PageList": ...

    @overload
    def __getitem__(self, key: Iterable[int]) -> "PageList": ...

    @overload
    def __getitem__(self, key: Iterator[Union[int, str]]) -> "PageList": ...

    def __getitem__(
        self, key: Union[int, str, slice, Iterable[int], Iterator[Union[int, str]]]
    ) -> Union[Page, "PageList"]:
        if isinstance(key, int):
            if self._pages is None:
                self._pages, self._by_label, self._by_objid = self._init_pages()
            return self._pages[key]
        elif isinstance(key, str):
            if self._by_label is None:
                self._pages, self._by_label, self._by_objid = self._init_pages()
            return self._by_label[key]
        elif isinstance(key, slice):
            if self._pages is None:
                self._pages, self._by_label, self._by_objid = self._init_pages()
            return PageList(_deref_document(self.docref), self._pages[key])
        else:
            return PageList(_deref_document(self.docref), (self[k] for k in key))

    def by_id(self, objid: int) -> Page:
        """Get a page by its indirect object ID.

        Args:
            objid: Indirect object ID for the page object.

        Returns:
            the page in question.
        """
        if self._by_objid is None:
            self._pages, self._by_label, self._by_objid = self._init_pages()
        return self._by_objid[objid]

    def map(self, func: Callable[[Page], Any]) -> Iterator:
        """Apply a function over each page, iterating over its results.

        Args:
            func: The function to apply to each page.

        Note:
            This possibly runs `func` in a separate process.  If its
            return value is not serializable (by `pickle`) then you
            will encounter errors.
        """
        doc = _deref_document(self.docref)
        if doc._pool is not None and (
            self._pages is None or len(self._pages) > doc._ncpus * doc._batchsize
        ):
            return doc._pool.map(
                call_page,
                itertools.repeat(func),
                ((id(doc), page.page_idx) for page in self),
            )
        else:
            return (func(page) for page in self)


def _iter_labels(tree: NumberTree) -> Iterator[str]:
    """Iterate over page label tree.

    See Section 12.4.2 in the PDF 1.7 Reference.
    """
    itor = iter(tree.items())
    try:
        start, label_dict_unchecked = next(itor)
        # The tree must begin with page index 0
        if start != 0:
            log.warning("Page label tree is missing index 0")
            # Try to cope, by assuming empty labels for the initial pages
            start = 0
    except StopIteration:
        log.warning("Page label tree is empty")
        start = 0
        label_dict_unchecked = {}

    while True:  # forever!
        label_dict = dict_value(label_dict_unchecked)
        style = label_dict.get("S")
        prefix = decode_text(str_value(label_dict.get("P", b"")))
        first_value = int_value(label_dict.get("St", 1))

        try:
            next_start, label_dict_unchecked = next(itor)
        except StopIteration:
            # This is the last specified range. It continues until the end
            # of the document.
            values: Iterable[int] = itertools.count(first_value)
        else:
            range_length = next_start - start
            values = range(first_value, first_value + range_length)
            start = next_start

        for value in values:
            label = _format_page_label(value, style)
            yield prefix + label


def _format_page_label(value: int, style: Any) -> str:
    """Format page label value in a specific style"""
    if style is None:
        label = ""
    elif style is LIT("D"):  # Decimal arabic numerals
        label = str(value)
    elif style is LIT("R"):  # Uppercase roman numerals
        label = format_int_roman(value).upper()
    elif style is LIT("r"):  # Lowercase roman numerals
        label = format_int_roman(value)
    elif style is LIT("A"):  # Uppercase letters A-Z, AA-ZZ...
        label = format_int_alpha(value).upper()
    elif style is LIT("a"):  # Lowercase letters a-z, aa-zz...
        label = format_int_alpha(value)
    else:
        log.warning("Unknown page label style: %r", style)
        label = ""
    return label


class Destinations(Mapping[Union[str, bytes, PSLiteral], Destination]):
    """Mapping of named destinations.

    These either come as a NameTree or a dict, depending on the
    version of the PDF standard, so this abstracts that away.
    """

    dests_dict: Union[Dict[str, PDFObject], None] = None
    dests_tree: Union[NameTree, None] = None

    def __init__(self, doc: Document) -> None:
        self._docref = _ref_document(doc)
        self.dests: Dict[str, Destination] = {}
        if "Dests" in doc.catalog:
            # PDF-1.1: dictionary
            dests_dict = resolve1(doc.catalog["Dests"])
            if isinstance(dests_dict, dict):
                self.dests_dict = dests_dict
            else:
                log.warning(
                    "Dests entry in catalog is not dictionary: %r", self.dests_dict
                )
                self.dests_dict = None
        elif "Names" in doc.catalog:
            names = resolve1(doc.catalog["Names"])
            if not isinstance(names, dict):
                log.warning("Names entry in catalog is not dictionary: %r", names)
                return
            if "Dests" in names:
                dests = resolve1(names["Dests"])
                if not isinstance(names, dict):
                    log.warning("Dests entry in names is not dictionary: %r", dests)
                    return
                self.dests_tree = NameTree(dests)

    def __len__(self) -> int:
        if self.dests_dict is not None:
            return len(self.dests_dict)
        if self.dests_tree is not None:
            return len(self.dests_tree)
        return 0

    def __iter__(self) -> Iterator[str]:
        """Iterate over names of destinations.

        Danger: Beware of corrupted PDFs
            This simply iterates over the names listed in the PDF, and
            does not attempt to actually parse the destinations
            (because that's pretty slow).  If the PDF is broken, you
            may encounter exceptions when actually trying to access
            them by name.
        """
        if self.dests_dict is not None:
            yield from self.dests_dict
        elif self.dests_tree is not None:
            for kb in self.dests_tree:
                ks = decode_text(kb)
                yield ks

    def items(self) -> "DestinationsItemsView":
        """Iterate over named destinations."""
        return DestinationsItemsView(self)

    def __getitem__(self, name: Union[bytes, str, PSLiteral]) -> Destination:
        """Get a named destination.

        Args:
            name: The name of the destination.

        Raises:
            KeyError: If no such destination exists.
            TypeError: If the PDF is damaged and the destinations tree
                contains something unexpected or missing.
        """
        if isinstance(name, bytes):
            name = decode_text(name)
        elif isinstance(name, PSLiteral):
            name = literal_name(name)
        if name in self.dests:
            return self.dests[name]
        elif self.dests_dict is not None:
            # This will raise KeyError or TypeError if necessary, so
            # we don't have to do it explicitly
            dest = resolve1(self.dests_dict[name])
            self.dests[name] = self._create_dest(dest, name)
        elif self.dests_tree is not None:
            # This is not at all efficient, but we need to decode
            # the keys (and we cache the result...)
            for k, v in self.dests_tree.items():
                if decode_text(k) == name:
                    dest = resolve1(v)
                    self.dests[name] = self._create_dest(dest, name)
                    break
        # This will also raise KeyError if necessary
        return self.dests[name]

    def _create_dest(self, dest: PDFObject, name: str) -> Destination:
        if isinstance(dest, list):
            return Destination.from_list(self.doc, dest)
        elif isinstance(dest, dict) and "D" in dest:
            destlist = resolve1(dest["D"])
            if not isinstance(destlist, list):
                raise TypeError("Invalid destination for %s: %r", name, dest)
            return Destination.from_list(self.doc, destlist)
        else:
            raise TypeError("Invalid destination for %s: %r", name, dest)

    @property
    def doc(self) -> "Document":
        """Get associated document if it exists."""
        return _deref_document(self._docref)


class DestinationsItemsView(ItemsView[str, Destination]):
    _mapping: Destinations

    def __iter__(self) -> Iterator[Tuple[str, Destination]]:
        dests = self._mapping
        if dests.dests_dict is not None:
            for name, dest in dests.dests_dict.items():
                if name not in dests.dests:
                    dest = resolve1(dests.dests_dict[name])
                    dests.dests[name] = dests._create_dest(dest, name)
                yield name, dests.dests[name]
        elif dests.dests_tree is not None:
            for k, v in dests.dests_tree.items():
                name = decode_text(k)
                if name not in dests.dests:
                    dest = resolve1(v)
                    dests.dests[name] = dests._create_dest(dest, name)
                yield name, dests.dests[name]



"""(Re-)create the JSON schemas for playa.data classes."""

import json
from pathlib import Path

from pydantic import TypeAdapter
from typing_extensions import is_typeddict

import playa.data

SCHEMA = Path(playa.data.__file__).parent / "schema"


def main():
    SCHEMA.mkdir(parents=True, exist_ok=True)
    for name, obj in vars(playa.data).items():
        if is_typeddict(obj):
            adapter = TypeAdapter(obj)
            with open(SCHEMA / f"{name}.json", "wt") as outfh:
                json.dump(adapter.json_schema(), outfh, indent=2, ensure_ascii=False)


if __name__ == "__main__":
    main()



"""
Benchmark the converter on all of the sample documents.
"""

import logging
import time
from pathlib import Path

import playa
from playa import ContentObject

LOG = logging.getLogger("benchmark-convert")
# Use a standard benchmark set to make version comparisons possible
CONTRIB = Path(__file__).parent.parent / "samples" / "contrib"
PDFS = [
    "PSC_Station.pdf",
    "Rgl-1314-2021-DM-Derogations-mineures.pdf",
]


def benchmark_one_lazy(path: Path):
    """Open one of the documents"""
    with playa.open(path) as pdf:
        for page in pdf.pages:
            obj: ContentObject
            for obj in page.texts:
                _ = obj.bbox
            for obj in page.paths:
                _ = obj.bbox
            for obj in page.images:
                _ = obj.bbox
            for obj in page.xobjects:
                _ = obj.bbox


if __name__ == "__main__":
    # Silence warnings about broken PDFs
    logging.basicConfig(level=logging.ERROR)
    niter = 5
    lazy_time = 0.0
    for iter in range(niter + 1):
        for name in PDFS:
            path = CONTRIB / name
            start = time.time()
            benchmark_one_lazy(path)
            if iter != 0:
                lazy_time += time.time() - start
    print("Object types took %d ms / iter" % (lazy_time / niter * 1000,))



"""Serializers for PDF metadata."""

import base64
import dataclasses
import functools
from typing import TypeVar, Union

from playa.parser import PSLiteral
from playa.pdftypes import ObjRef, literal_name


@functools.singledispatch
def asobj(obj):
    """JSON serializable representation of PDF object metadata."""
    # functools.singledispatch can't register None
    if obj is None:
        return None
    # Catch dataclasses that don't have a specific serializer
    if dataclasses.is_dataclass(obj) and not isinstance(obj, type):
        return {k: asobj(v) for k, v in obj.__dict__.items()}
    return repr(obj)


_S = TypeVar("_S", int, float, bool, str)


def asobj_simple(obj: _S) -> _S:
    return obj


# Have to list these all for Python <3.11 where
# functools.singledispatch doesn't support Union
asobj.register(int, asobj_simple)
asobj.register(float, asobj_simple)
asobj.register(bool, asobj_simple)
asobj.register(str, asobj_simple)


@asobj.register
def asobj_bytes(obj: bytes) -> str:
    # Reimplement decode_text here as we want to be stricter about
    # what we consider a text string.  PDFDocEncoding is impossible to
    # detect so should only be used when we *know* it's a text string
    # according to the PDF standard.
    try:
        if obj.startswith(b"\xfe\xff") or obj.startswith(b"\xff\xfe"):
            return obj.decode("UTF-16")
        return obj.decode("ascii")
    except UnicodeDecodeError:
        # FIXME: This may be subject to change...
        return "base64:" + base64.b64encode(obj).decode("ascii")


@asobj.register
def asobj_literal(obj: PSLiteral) -> str:
    return literal_name(obj)


@asobj.register
def asobj_dict(obj: dict) -> dict:
    return {k: asobj(v) for k, v in obj.items()}


@asobj.register
def asobj_list(obj: list) -> list:
    return [asobj(v) for v in obj]


@asobj.register(tuple)
def asobj_tuple(obj: tuple) -> Union[list, dict]:
    # Catch NamedTuples that don't have a specific serializer
    if hasattr(obj, "_asdict"):
        return {k: asobj(v) for k, v in obj._asdict().items()}
    return [asobj(v) for v in obj]


@asobj.register
def asobj_ref(obj: ObjRef) -> str:
    # This is the same as repr() but we want it defined separately
    return f"<ObjRef:{obj.objid}>"



"""
Page-related tests
"""

import pytest

import playa
from .data import TESTDIR


def test_rotation() -> None:
    datadir = TESTDIR / "rotation"
    with playa.open(datadir / "0.pdf") as pdf:
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx((100, 74.768, 224.008, 96.968))
    with playa.open(datadir / "90.pdf") as pdf:
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx((792 - 96.968, 100, 792 - 74.768, 224.008))
    with playa.open(datadir / "180.pdf") as pdf:
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx(
            (612 - 224.008, 792 - 96.968, 612 - 100, 792 - 74.768)
        )
    with playa.open(datadir / "270.pdf") as pdf:
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx((74.768, 612 - 224.008, 96.968, 612 - 100))
        # And verify that we can update the rotation
        pdf.pages[0].set_initial_ctm("screen", 0)
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx((100, 74.768, 224.008, 96.968))
        # And verify that rotation is normalized
        pdf.pages[0].set_initial_ctm("screen", -90)
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx((74.768, 612 - 224.008, 96.968, 612 - 100))


def test_translation() -> None:
    datadir = TESTDIR / "rotation"
    with playa.open(datadir / "0mb.pdf") as pdf:
        hello = next(pdf.pages[0].texts)
        # screen device space so no effect on Y coords
        assert hello.bbox == pytest.approx((90, 74.768, 214.008, 96.968))
    with playa.open(datadir / "0.pdf", space="page") as pdf:
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx((100, 695.032, 224.008, 717.232))
    with playa.open(datadir / "0mb.pdf", space="page") as pdf:
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx((90, 675.032, 214.008, 697.232))
    with playa.open(datadir / "90mb.pdf") as pdf:
        hello = next(pdf.pages[0].texts)
        # Just trust me, these are correct
        assert hello.bbox == pytest.approx((675.032, 90, 697.232, 214.008))
    with playa.open(datadir / "180mb.pdf") as pdf:
        hello = next(pdf.pages[0].texts)
        # Just trust me, these are correct
        assert hello.bbox == pytest.approx((387.992, 675.032, 512, 697.232))
    with playa.open(datadir / "270mb.pdf") as pdf:
        hello = next(pdf.pages[0].texts)
        # No change here because we cropped the other corner
        assert hello.bbox == pytest.approx((74.768, 612 - 224.008, 96.968, 612 - 100))


def test_set_initial_ctm(caplog) -> None:
    datadir = TESTDIR / "rotation"
    with playa.open(datadir / "0.pdf") as pdf:
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx((100, 74.768, 224.008, 96.968))
        pdf.pages[0].set_initial_ctm("default", 0)
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx((100, 792 - 96.968, 224.008, 792 - 74.768))
        # Rotation should have no effect on "default" (but it will set
        # the rotate property anyway)
        pdf.pages[0].set_initial_ctm("default", 90)
        hello = next(pdf.pages[0].texts)
        assert pdf.pages[0].rotate == 90
        assert hello.bbox == pytest.approx((100, 792 - 96.968, 224.008, 792 - 74.768))
        # Other spaces should do the same thing as "default"
        pdf.pages[0].set_initial_ctm("spam", 180)  # type: ignore[arg-type]
        assert "Unknown device space" in caplog.text
        hello = next(pdf.pages[0].texts)
        assert pdf.pages[0].rotate == 180
        assert hello.bbox == pytest.approx((100, 792 - 96.968, 224.008, 792 - 74.768))
        # Non-90 degree rotation defaults to 0
        pdf.pages[0].set_initial_ctm("screen", 42)
        assert "Invalid rotation" in caplog.text
        hello = next(pdf.pages[0].texts)
        assert hello.bbox == pytest.approx((100, 74.768, 224.008, 96.968))



"""Miscellaneous Routines."""

import string
from typing import (
    TYPE_CHECKING,
    Final,
    Iterable,
    Iterator,
    List,
    Mapping,
    Tuple,
    TypeVar,
    Union,
)

from playa.pdftypes import PDFObject, Point, Rect, Matrix, dict_value, str_value

if TYPE_CHECKING:
    from mypy_extensions import u8
    from playa.content import GraphicState
    from playa.font import Font


def paeth_predictor(left: int, above: int, upper_left: int) -> int:
    # From http://www.libpng.org/pub/png/spec/1.2/PNG-Filters.html
    # Initial estimate
    p = left + above - upper_left
    # Distances to a,b,c
    pa = abs(p - left)
    pb = abs(p - above)
    pc = abs(p - upper_left)

    # Return nearest of a,b,c breaking ties in order a,b,c
    if pa <= pb and pa <= pc:
        return left
    elif pb <= pc:
        return above
    else:
        return upper_left


def apply_tiff_predictor(
    colors: int, columns: int, bitspercomponent: int, data: bytes
) -> bytes:
    """Reverse the effect of the TIFF predictor 2

    Documentation: https://www.itu.int/itudoc/itu-t/com16/tiff-fx/docs/tiff6.pdf
    (Section 14, page 64)
    """
    if bitspercomponent != 8:
        error_msg = f"Unsupported `bitspercomponent': {bitspercomponent}"
        raise ValueError(error_msg)
    bpp = colors * (bitspercomponent // 8)
    nbytes = columns * bpp
    # Incredibly, this is faster than bytearray (with and without mypyc)
    buf: List["u8"] = []
    for scanline_i in range(0, len(data), nbytes):
        for i in range(nbytes):
            new_value = data[scanline_i + i]
            if i >= bpp:
                new_value = (new_value + buf[scanline_i + i - bpp]) & 255
            buf.append(new_value)
    return bytes(buf)


def apply_png_predictor(
    pred: int,
    colors: int,
    columns: int,
    bitspercomponent: int,
    data: bytes,
) -> bytes:
    """Reverse the effect of the PNG predictor

    Documentation: http://www.libpng.org/pub/png/spec/1.2/PNG-Filters.html
    """
    if bitspercomponent not in (8, 1):
        msg = "Unsupported `bitspercomponent': %d" % bitspercomponent
        raise ValueError(msg)

    nbytes = (colors * columns * bitspercomponent + 7) // 8
    bpp = (colors * bitspercomponent + 7) // 8
    # Likewise, this is *twice as fast* as bytearray...
    buf: List["u8"] = []
    line_above: List["u8"] = [0] * nbytes
    raw: List["u8"] = [0] * nbytes
    for scanline_i in range(0, len(data), nbytes + 1):
        filter_type = data[scanline_i]
        line_encoded = data[scanline_i + 1 : scanline_i + 1 + nbytes]

        if filter_type == 0:
            # Filter type 0: None
            raw[:] = line_encoded

        elif filter_type == 1:
            # Filter type 1: Sub
            # To reverse the effect of the Sub() filter after decompression,
            # output the following value:
            #   Raw(x) = Sub(x) + Raw(x - bpp)
            # (computed mod 256), where Raw() refers to the bytes already
            #  decoded.
            for j, sub_x in enumerate(line_encoded):
                if j < bpp:
                    raw[j] = sub_x
                else:
                    raw[j] = (sub_x + raw[j - bpp]) & 255

        elif filter_type == 2:
            # Filter type 2: Up
            # To reverse the effect of the Up() filter after decompression,
            # output the following value:
            #   Raw(x) = Up(x) + Prior(x)
            # (computed mod 256), where Prior() refers to the decoded bytes of
            # the prior scanline.
            for j, up_x in enumerate(line_encoded):
                prior_x = line_above[j]
                raw[j] = (up_x + prior_x) & 255

        elif filter_type == 3:
            # Filter type 3: Average
            # To reverse the effect of the Average() filter after
            # decompression, output the following value:
            #    Raw(x) = Average(x) + floor((Raw(x-bpp)+Prior(x))/2)
            # where the result is computed mod 256, but the prediction is
            # calculated in the same way as for encoding. Raw() refers to the
            # bytes already decoded, and Prior() refers to the decoded bytes of
            # the prior scanline.
            for j, average_x in enumerate(line_encoded):
                prior_x = line_above[j]
                if j < bpp:
                    raw_x_bpp = 0
                else:
                    raw_x_bpp = raw[j - bpp]
                raw[j] = (average_x + (raw_x_bpp + prior_x) // 2) & 255

        elif filter_type == 4:
            # Filter type 4: Paeth
            # To reverse the effect of the Paeth() filter after decompression,
            # output the following value:
            #    Raw(x) = Paeth(x)
            #             + PaethPredictor(Raw(x-bpp), Prior(x), Prior(x-bpp))
            # (computed mod 256), where Raw() and Prior() refer to bytes
            # already decoded. Exactly the same PaethPredictor() function is
            # used by both encoder and decoder.
            for j, paeth_x in enumerate(line_encoded):
                if j - bpp < 0:
                    raw_x_bpp = 0
                    prior_x_bpp = 0
                else:
                    raw_x_bpp = raw[j - bpp]
                    prior_x_bpp = line_above[j - bpp]
                prior_x = line_above[j]
                paeth = paeth_predictor(raw_x_bpp, prior_x, prior_x_bpp)
                raw[j] = (paeth_x + paeth) & 255

        else:
            raise ValueError("Unsupported predictor value: %d" % filter_type)

        buf.extend(raw)
        raw, line_above = line_above, raw
    return bytes(buf)


def normalize_rect(r: Rect) -> Rect:
    (x0, y0, x1, y1) = r
    if x1 < x0:
        x1, x0 = x0, x1
    if y1 < y0:
        y1, y0 = y0, y1
    return x0, y0, x1, y1


def intersect_rects(a: Rect, b: Rect) -> Rect:
    (ax0, ay0, ax1, ay1) = normalize_rect(a)
    (bx0, by0, bx1, by1) = normalize_rect(b)
    x0 = max(ax0, bx0)
    x1 = min(ax1, bx1)
    y0 = max(ay0, by0)
    y1 = min(ay1, by1)
    return x0, y0, x1, y1


def point_in_rect(p: Point, r: Rect) -> bool:
    x, y = p
    x0, y0, x1, y1 = normalize_rect(r)
    return x0 <= x <= x1 and y0 <= y <= y1


def mult_matrix(m1: Matrix, m0: Matrix) -> Matrix:
    (a1, b1, c1, d1, e1, f1) = m1
    (a0, b0, c0, d0, e0, f0) = m0
    """Returns the multiplication of two matrices."""
    return (
        a0 * a1 + c0 * b1,
        b0 * a1 + d0 * b1,
        a0 * c1 + c0 * d1,
        b0 * c1 + d0 * d1,
        a0 * e1 + c0 * f1 + e0,
        b0 * e1 + d0 * f1 + f0,
    )


def translate_matrix(m: Matrix, v: Point) -> Matrix:
    """Pre-translates a matrix by `v == (dx, dy)`.

    Specifically this translates the *input space* of the matrix by
    `(dx, dy)`, so:

       (apply_matrix_pt(matrix, (x + dx, y + dy))
        == apply_matrix_pt(translate_matrix(matrix,
                                            (dx, dy)),
                           x, y))

    For all `x, y, dx, dy`.
    """
    (a, b, c, d, e, f) = m
    (x, y) = v
    return a, b, c, d, x * a + y * c + e, x * b + y * d + f


def apply_matrix_pt(m: Matrix, v: Point) -> Point:
    """Applies a matrix to a point."""
    (a, b, c, d, e, f) = m
    (x, y) = v
    return a * x + c * y + e, b * x + d * y + f


def apply_matrix_norm(m: Matrix, v: Point) -> Point:
    """Equivalent to apply_matrix_pt(M, (p,q)) - apply_matrix_pt(M, (0,0))"""
    (a, b, c, d, e, f) = m
    (p, q) = v
    return a * p + c * q, b * p + d * q


#  Utility functions
_T = TypeVar("_T")


def get_bound(pts: Iterable[Point]) -> Rect:
    """Compute a minimal rectangle that covers all the points.

    Raises:
      ValueError: on empty input (as there is no bounding box).
    """
    xs, ys = zip(*pts)
    x0 = min(xs)
    y0 = min(ys)
    x1 = max(xs)
    y1 = max(ys)
    return x0, y0, x1, y1


def get_bound_rects(boxes: Iterable[Union[Rect, None]]) -> Union[Rect, None]:
    """Compute the union of bounding boxes (which need not be normalized)."""
    points = []
    for box in boxes:
        if box is None:
            continue
        x0, y0, x1, y1 = box
        points.append((x0, y0))
        points.append((x1, y1))
    if points:
        return get_bound(points)
    return None


def transform_bbox(matrix: Matrix, bbox: Rect) -> Rect:
    """Transform a bounding box and return the rectangle that covers
    the points of the resulting shape."""
    x0, y0, x1, y1 = bbox
    if matrix[0] * matrix[2] < 0 or matrix[1] * matrix[3] < 0:
        # Corners of untransformed bbox are not preserved
        return get_bound(
            (
                apply_matrix_pt(matrix, (x0, y0)),
                apply_matrix_pt(matrix, (x0, y1)),
                apply_matrix_pt(matrix, (x1, y1)),
                apply_matrix_pt(matrix, (x1, y0)),
            )
        )
    return get_bound(
        (
            apply_matrix_pt(matrix, (x0, y0)),
            apply_matrix_pt(matrix, (x1, y1)),
        )
    )


def choplist(n: int, seq: Iterable[_T]) -> Iterator[Tuple[_T, ...]]:
    """Groups every n elements of the list."""
    r = []
    for x in seq:
        r.append(x)
        if len(r) == n:
            yield tuple(r)
            r = []


def nunpack(s: bytes, default: int = 0) -> int:
    """Unpacks variable-length unsigned integers (big endian)."""
    length = len(s)
    if not length:
        return default
    else:
        return int.from_bytes(s, byteorder="big", signed=False)


PDFDocEncoding: Final[str] = "".join(
    chr(x)
    for x in (
        0x0000,
        0x0001,
        0x0002,
        0x0003,
        0x0004,
        0x0005,
        0x0006,
        0x0007,
        0x0008,
        0x0009,
        0x000A,
        0x000B,
        0x000C,
        0x000D,
        0x000E,
        0x000F,
        0x0010,
        0x0011,
        0x0012,
        0x0013,
        0x0014,
        0x0015,
        0x0017,
        0x0017,
        0x02D8,
        0x02C7,
        0x02C6,
        0x02D9,
        0x02DD,
        0x02DB,
        0x02DA,
        0x02DC,
        0x0020,
        0x0021,
        0x0022,
        0x0023,
        0x0024,
        0x0025,
        0x0026,
        0x0027,
        0x0028,
        0x0029,
        0x002A,
        0x002B,
        0x002C,
        0x002D,
        0x002E,
        0x002F,
        0x0030,
        0x0031,
        0x0032,
        0x0033,
        0x0034,
        0x0035,
        0x0036,
        0x0037,
        0x0038,
        0x0039,
        0x003A,
        0x003B,
        0x003C,
        0x003D,
        0x003E,
        0x003F,
        0x0040,
        0x0041,
        0x0042,
        0x0043,
        0x0044,
        0x0045,
        0x0046,
        0x0047,
        0x0048,
        0x0049,
        0x004A,
        0x004B,
        0x004C,
        0x004D,
        0x004E,
        0x004F,
        0x0050,
        0x0051,
        0x0052,
        0x0053,
        0x0054,
        0x0055,
        0x0056,
        0x0057,
        0x0058,
        0x0059,
        0x005A,
        0x005B,
        0x005C,
        0x005D,
        0x005E,
        0x005F,
        0x0060,
        0x0061,
        0x0062,
        0x0063,
        0x0064,
        0x0065,
        0x0066,
        0x0067,
        0x0068,
        0x0069,
        0x006A,
        0x006B,
        0x006C,
        0x006D,
        0x006E,
        0x006F,
        0x0070,
        0x0071,
        0x0072,
        0x0073,
        0x0074,
        0x0075,
        0x0076,
        0x0077,
        0x0078,
        0x0079,
        0x007A,
        0x007B,
        0x007C,
        0x007D,
        0x007E,
        0x0000,
        0x2022,
        0x2020,
        0x2021,
        0x2026,
        0x2014,
        0x2013,
        0x0192,
        0x2044,
        0x2039,
        0x203A,
        0x2212,
        0x2030,
        0x201E,
        0x201C,
        0x201D,
        0x2018,
        0x2019,
        0x201A,
        0x2122,
        0xFB01,
        0xFB02,
        0x0141,
        0x0152,
        0x0160,
        0x0178,
        0x017D,
        0x0131,
        0x0142,
        0x0153,
        0x0161,
        0x017E,
        0x0000,
        0x20AC,
        0x00A1,
        0x00A2,
        0x00A3,
        0x00A4,
        0x00A5,
        0x00A6,
        0x00A7,
        0x00A8,
        0x00A9,
        0x00AA,
        0x00AB,
        0x00AC,
        0x0000,
        0x00AE,
        0x00AF,
        0x00B0,
        0x00B1,
        0x00B2,
        0x00B3,
        0x00B4,
        0x00B5,
        0x00B6,
        0x00B7,
        0x00B8,
        0x00B9,
        0x00BA,
        0x00BB,
        0x00BC,
        0x00BD,
        0x00BE,
        0x00BF,
        0x00C0,
        0x00C1,
        0x00C2,
        0x00C3,
        0x00C4,
        0x00C5,
        0x00C6,
        0x00C7,
        0x00C8,
        0x00C9,
        0x00CA,
        0x00CB,
        0x00CC,
        0x00CD,
        0x00CE,
        0x00CF,
        0x00D0,
        0x00D1,
        0x00D2,
        0x00D3,
        0x00D4,
        0x00D5,
        0x00D6,
        0x00D7,
        0x00D8,
        0x00D9,
        0x00DA,
        0x00DB,
        0x00DC,
        0x00DD,
        0x00DE,
        0x00DF,
        0x00E0,
        0x00E1,
        0x00E2,
        0x00E3,
        0x00E4,
        0x00E5,
        0x00E6,
        0x00E7,
        0x00E8,
        0x00E9,
        0x00EA,
        0x00EB,
        0x00EC,
        0x00ED,
        0x00EE,
        0x00EF,
        0x00F0,
        0x00F1,
        0x00F2,
        0x00F3,
        0x00F4,
        0x00F5,
        0x00F6,
        0x00F7,
        0x00F8,
        0x00F9,
        0x00FA,
        0x00FB,
        0x00FC,
        0x00FD,
        0x00FE,
        0x00FF,
    )
)


def decode_text(s: bytes) -> str:
    """Decodes a text string (see PDF 1.7 section 7.9.2.2 - it could
    be PDFDocEncoding or UTF-16BE) to a `str`.
    """
    # Sure, it could be UTF-16LE... \/\/hatever...
    if isinstance(s, bytes) and (
        s.startswith(b"\xfe\xff") or s.startswith(b"\xff\xfe")
    ):
        try:
            return s.decode("UTF-16")
        except UnicodeDecodeError:
            # Sure, it could have a BOM and not actually be UTF-16, \/\/TF...
            s = s[2:]
    try:
        return "".join(PDFDocEncoding[c] for c in s)
    except IndexError:
        # This is obviously wrong, but a reasonable fallback
        return s.decode("iso-8859-1")


def string_property(obj: PDFObject, key: str) -> Union[str, None]:
    """Retrieve and decode a string property of an object dictionary
    (such as an annotation, a page, a structure element, etc)

    Takes care of resolving indirect object references in both the
    object and its values.

    Raises:
        TypeError: If `obj` is not a dictionary, or if the value
        requested is not a string

    Returns:
        The decoded string, or `None` if the property doesn't exist.
    """
    objdict = dict_value(obj)
    val = objdict.get(key)
    if val is None:
        return None
    val = str_value(val)
    return decode_text(val)


ROMAN_ONES: Final = ["i", "x", "c", "m"]
ROMAN_FIVES: Final = ["v", "l", "d"]


def format_int_roman(value: int) -> str:
    """Format a number as lowercase Roman numerals."""
    if value <= 0 or value >= 4000:
        raise ValueError(f"Romanes eunt domus: {value}")
    result: List[str] = []
    index = 0

    while value != 0:
        value, remainder = divmod(value, 10)
        if remainder == 9:
            result.insert(0, ROMAN_ONES[index])
            result.insert(1, ROMAN_ONES[index + 1])
        elif remainder == 4:
            result.insert(0, ROMAN_ONES[index])
            result.insert(1, ROMAN_FIVES[index])
        else:
            over_five = remainder >= 5
            if over_five:
                result.insert(0, ROMAN_FIVES[index])
                remainder -= 5
            result.insert(1 if over_five else 0, ROMAN_ONES[index] * remainder)
        index += 1

    return "".join(result)


def format_int_alpha(value: int) -> str:
    """Format a number as lowercase letters a-z, aa-zz, etc."""
    if value <= 0:
        raise ValueError(f"No alphabetic page number for {value}")
    result: List[str] = []

    while value != 0:
        value, remainder = divmod(value - 1, len(string.ascii_lowercase))
        result.append(string.ascii_lowercase[remainder])

    result.reverse()
    return "".join(result)


class IdentityMapping(Mapping[_T, _T]):
    def __getitem__(self, key: _T) -> _T:
        return key

    def __iter__(self) -> Iterator[_T]:
        yield from ()

    def __len__(self) -> int:
        return 0


IDENTITY_MAPPING: IdentityMapping = IdentityMapping()


def update_glyph_offset(
    origin: Point,
    font: "Font",
    gstate: "GraphicState",
    args: List[Union[bytes, int, float]],
) -> Point:
    fontsize = gstate.fontsize
    horizontal_scaling = gstate.scaling * 0.01
    charspace = gstate.charspace
    wordspace = gstate.wordspace
    vert = font.vertical
    if font.multibyte:
        wordspace = 0.0
    (x, y) = origin
    if vert:
        return x, update_glyph_offset_vertical(
            y, font, args, vscale=fontsize, charspace=charspace, wordspace=wordspace
        )
    else:
        return (
            update_glyph_offset_horizontal(
                x,
                font,
                args,
                hscale=fontsize * horizontal_scaling,
                charspace=charspace * horizontal_scaling,
                wordspace=wordspace * horizontal_scaling,
            ),
            y,
        )


def update_glyph_offset_horizontal(
    x: float,
    font: "Font",
    args: List[Union[bytes, int, float]],
    hscale: float,
    charspace: float,
    wordspace: float,
) -> float:
    for obj in args:
        if isinstance(obj, (int, float)):
            x -= obj * 0.001 * hscale
        else:
            for cid, _ in font.decode(obj):
                x += font.hdisp(cid) * hscale + charspace
                if cid == 32:
                    x += wordspace
    return x


def update_glyph_offset_vertical(
    y: float,
    font: "Font",
    args: List[Union[bytes, int, float]],
    vscale: float,
    charspace: float,
    wordspace: float,
) -> float:
    for obj in args:
        if isinstance(obj, (int, float)):
            y -= obj * 0.001 * vscale
        else:
            for cid, _ in font.decode(obj):
                y += font.vdisp(cid) * vscale + charspace
                if cid == 32:
                    y += wordspace
    return y



"""Test parallel analysis."""

import operator
from typing import Dict, List

import playa
import pytest
from playa.pdftypes import PDFObject, PSLiteral, PSKeyword
from playa.page import Page, XObjectObject
from playa.worker import (
    _deref_document,
    _deref_page,
    _get_document,
    _init_worker,
    _init_worker_buffer,
    _ref_document,
    _ref_page,
    _set_document,
    in_worker,
)

from tests.data import CONTRIB, TESTDIR


def has_one_true_pdf() -> int:
    assert in_worker()
    doc = _get_document()
    assert doc is not None
    assert doc.space == "default"
    return len(doc.pages)


def test_open_parallel():
    with playa.open(
        TESTDIR / "pdf_structure.pdf", space="default", max_workers=4
    ) as pdf:
        future = pdf._pool.submit(has_one_true_pdf)
        assert future.result() == 1
    with playa.open(
        TESTDIR / "pdf_structure.pdf", space="default", max_workers=None
    ) as pdf:
        future = pdf._pool.submit(has_one_true_pdf)
        assert future.result() == 1


def test_parse_parallel():
    with open(TESTDIR / "pdf_structure.pdf", "rb") as infh:
        buffer = infh.read()
    with playa.parse(buffer, space="default", max_workers=4) as pdf:
        future = pdf._pool.submit(has_one_true_pdf)
        assert future.result() == 1
    with playa.parse(buffer, space="default", max_workers=None) as pdf:
        future = pdf._pool.submit(has_one_true_pdf)
        assert future.result() == 1


def test_parallel_references():
    with playa.open(
        TESTDIR / "pdf_structure.pdf", space="default", max_workers=2
    ) as pdf:
        # Force parallelism
        pdf._ncpus = 0
        (resources,) = pdf.pages.map(operator.attrgetter("resources"))
        desc = resources["Font"].resolve()  # should succeed!
        assert "F1" in desc  # should exist!
        assert "F2" in desc
        assert desc["F1"].resolve()["LastChar"] == 17


def get_props(page: Page) -> Dict[str, PDFObject]:
    element = page.structure[0]
    return {} if element is None else element.props


def get_contents(page: Page) -> List[PDFObject]:
    return list(page.contents)


def test_parallel_symbols():
    """Verify that literal/keyword symbols can be passed between processes."""
    with playa.open(
        TESTDIR / "pdf_structure.pdf", space="default", max_workers=2
    ) as pdf:
        # Force parallelism
        pdf._ncpus = 0
        props1 = [get_props(p) for p in pdf.pages]
        props2 = list(pdf.pages.map(get_props))
        assert isinstance(props1[0]["Type"], PSLiteral)
        assert props1[0]["Type"] is props2[0]["Type"]

        contents1 = [get_contents(p) for p in pdf.pages]
        contents2 = list(pdf.pages.map(get_contents))
        assert isinstance(contents1[0][3], PSKeyword)
        assert contents1[0][3] is contents2[0][3]


def get_xobjs(page: Page) -> List[XObjectObject]:
    return list(page.xobjects)


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_parallel_xobjects():
    # Verify that page references (used in XObjects) also work
    with playa.open(CONTRIB / "basicapi.pdf", space="default", max_workers=2) as pdf:
        # Force parallelism
        pdf._ncpus = 0
        for page in pdf.pages:
            for xobj in page.xobjects:
                assert xobj.page.page_idx == page.page_idx
        for idx, xobjs in enumerate(pdf.pages.map(get_xobjs)):
            for xobj in xobjs:
                assert xobj.page.page_idx == idx


def get_text(page: Page) -> str:
    return " ".join(x.chars for x in page.texts)


def in_worker_page(page: Page) -> int:
    return in_worker()


def test_map_not_parallel():
    """Verify that we will not execute in parallel for a single page."""
    # This document has only one page
    with playa.open(
        TESTDIR / "pdf_structure.pdf", space="default", max_workers=2
    ) as pdf:
        in_workers = list(pdf.pages.map(in_worker_page))
        assert in_workers == [False]


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_map_parallel():
    # Verify also that a longer document will execute in parallel!
    with playa.open(CONTRIB / "PSC_Station.pdf", space="default", max_workers=2) as pdf:
        in_workers = list(pdf.pages.map(in_worker_page))
        assert all(in_workers)
    with playa.open(CONTRIB / "PSC_Station.pdf", space="default", max_workers=2) as pdf:
        parallel_texts = list(pdf.pages.map(get_text))
    with playa.open(CONTRIB / "PSC_Station.pdf", space="default") as pdf:
        texts = list(pdf.pages.map(get_text))
    assert texts == parallel_texts
    with playa.open(CONTRIB / "PSC_Station.pdf", space="default", max_workers=2) as pdf:
        parallel_texts = list(pdf.pages[3:8].map(get_text))
        print(parallel_texts)
        assert parallel_texts != texts
    # Verify that a shorter slice will not execute in parallel
    with playa.open(CONTRIB / "PSC_Station.pdf", space="default", max_workers=2) as pdf:
        in_workers = list(pdf.pages[0:2].map(in_worker_page))
        assert not any(in_workers)


def test_worker():
    """Ensure coverage of worker functions (even though they are tested above)."""
    _init_worker(123456, TESTDIR / "pdf_structure.pdf", "", {})
    pdf1 = _get_document()
    assert pdf1
    assert in_worker()
    docref = _ref_document(pdf1)
    assert docref == 123456
    assert _deref_document(docref) is pdf1
    pageref = _ref_page(pdf1.pages[0])
    assert pageref == (123456, 0)
    assert _deref_page(pageref) is pdf1.pages[0]
    _set_document(None, 0)
    assert not in_worker()
    with pytest.raises(RuntimeError):
        _deref_document(docref)
    with playa.open(TESTDIR / "image_structure.pdf") as pdf2:
        _set_document(pdf2, 654321)
        assert _get_document() is pdf2
    with open(TESTDIR / "image_structure.pdf", "rb") as fh:
        _init_worker_buffer(654321, fh.read(), "", {})
        assert _get_document()
    _set_document(None, 0)
    pdf3 = playa.open(TESTDIR / "image_structure.pdf")
    docref = _ref_document(pdf3)
    assert _deref_document(docref) is pdf3
    del pdf3
    with pytest.raises(RuntimeError):
        _deref_document(docref)


if __name__ == "__main__":
    test_parallel_xobjects()



"""
Benchmark text extraction on some sample documents.
"""

import logging
import time
from pathlib import Path

import playa

SAMPLES = Path(__file__).parent.parent / "samples"
CONTRIB = Path(__file__).parent.parent / "samples" / "contrib"

LOG = logging.getLogger("benchmark-text")
# Use a standard benchmark set to make version comparisons possible
PDFS = [
    "jo.pdf",
    "zen_of_python_corrupted.pdf",
    "2023-04-06-ODJ et Résolutions-séance xtra 6 avril 2023.pdf",
    "2023-06-20-PV.pdf",
    "PSC_Station.pdf",
    "Rgl-1314-2021-DM-Derogations-mineures.pdf",
]


def benchmark_chars(path: Path):
    """Extract just the Unicode characters (a poor substitute for actual
    text extraction)"""

    with playa.open(path) as pdf:
        for page in pdf.pages:
            for obj in page.texts:
                _ = obj.chars


def benchmark_text(path: Path):
    """Extract text, sort of."""
    with playa.open(path) as pdf:
        for page in pdf.pages:
            page.extract_text()


if __name__ == "__main__":
    logging.basicConfig(level=logging.ERROR)
    niter = 5
    chars_time = text_time = 0.0
    for iter in range(niter + 1):
        for name in PDFS:
            path = SAMPLES / name
            if not path.exists():
                path = CONTRIB / name
            start = time.time()
            benchmark_chars(path)
            if iter != 0:
                chars_time += time.time() - start
            start = time.time()
            benchmark_text(path)
            if iter != 0:
                text_time += time.time() - start
    print("chars took %d ms / iter" % (chars_time / niter * 1000,))
    print("extract_text took %d ms / iter" % (text_time / niter * 1000,))



"""
Test the classes in data_structures.py
"""

from playa.data_structures import NameTree, NumberTree


def test_name_trees() -> None:
    """Test name trees."""
    nt1 = NameTree(
        {
            "Names": [b"arthur", 1, b"jackson", 3, b"sheds", 4, b"two", 2],
            "Limits": [b"arthur", b"two"],
            "Kids": [{"Names": [b"dead", 33, b"parrot", 66]}],
        }
    )
    assert nt1[b"parrot"] == 66
    assert b"urbanisme" not in nt1
    assert list(nt1) == [b"arthur", b"jackson", b"sheds", b"two", b"dead", b"parrot"]
    assert list(nt1.values()) == [1, 3, 4, 2, 33, 66]


def test_number_trees() -> None:
    """Test number trees."""
    nt1 = NumberTree(
        {
            "Nums": [0, 1, 2, 3, 10, 5],
            "Limits": [0, 10],
            "Kids": [{"Nums": [4, 5, 6, 7]}],
        }
    )
    assert nt1[4] == 5
    assert 11 not in nt1
    assert list(nt1) == [0, 2, 10, 4, 6]
    assert list(nt1.values()) == [1, 3, 5, 5, 7]


NUMTREE1 = {
    "Kids": [
        {"Nums": [1, b"a", 3, b"b", 7, b"c"], "Limits": [1, 7]},
        {
            "Kids": [
                {"Nums": [8, 123, 9, {"x": b"y"}, 10, b"forty-two"], "Limits": [8, 10]},
                {"Nums": [11, b"zzz", 12, b"xxx", 15, b"yyy"], "Limits": [11, 15]},
            ],
            "Limits": [8, 15],
        },
        {"Nums": [20, 456], "Limits": [20, 20]},
    ]
}


def test_number_tree():
    """Test NumberTrees."""
    nt = NumberTree(NUMTREE1)
    assert 15 in nt
    assert 20 in nt
    assert nt[20] == 456
    assert nt[9] == {"x": b"y"}
    assert list(nt.items()) == [
        (1, b"a"),
        (3, b"b"),
        (7, b"c"),
        (8, 123),
        (9, {"x": b"y"}),
        (10, b"forty-two"),
        (11, b"zzz"),
        (12, b"xxx"),
        (15, b"yyy"),
        (20, 456),
    ]
    assert (20, 456) in nt.items()
    assert 20 in nt.keys()
    assert 456 in nt.values()


NAMETREE1 = {
    "Kids": [
        {"Names": [b"bletch", b"a", b"foobie", b"b"], "Limits": [b"bletch", b"foobie"]},
        {
            "Kids": [
                {
                    "Names": [b"gargantua", 35, b"gorgon", 42],
                    "Limits": [b"gargantua", b"gorgon"],
                },
                {
                    "Names": [b"xylophone", 123, b"zzyzx", {"x": b"y"}],
                    "Limits": [b"xylophone", b"zzyzx"],
                },
            ],
            "Limits": [b"gargantua", b"zzyzx"],
        },
    ]
}


def test_name_tree():
    """Test NameTrees."""
    nt = NameTree(NAMETREE1)
    assert b"bletch" in nt
    assert b"zzyzx" in nt
    assert b"gorgon" in nt
    assert nt[b"zzyzx"] == {"x": b"y"}
    assert list(nt.items()) == [
        (b"bletch", b"a"),
        (b"foobie", b"b"),
        (b"gargantua", 35),
        (b"gorgon", 42),
        (b"xylophone", 123),
        (b"zzyzx", {"x": b"y"}),
    ]
    assert (b"gargantua", 35) in nt.items()
    assert b"gargantua" in nt.keys()
    assert 35 in nt.values()


if __name__ == "__main__":
    test_name_tree()
    test_number_tree()
